"""
The basic definitions for file-like reading and writing. This is generally
centered on image-like file efforts, and array-like interaction with image data.

This module completely revamped in version 1.3.0 for data segment usage.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import os
import logging
from typing import Union, List, Tuple, Sequence, Optional, Callable
from importlib import import_module
import pkgutil

import numpy

import sarpy._extensions
from sarpy.compliance import SarpyError
from sarpy.io.general.format_function import FormatFunction
from sarpy.io.general.data_segment import DataSegment, extract_string_from_subscript, \
    NumpyArraySegment

logger = logging.getLogger(__name__)

READER_TYPES = ('SICD', 'SIDD', 'CPHD', 'CRSD', 'OTHER')
"""
The reader_type enum 
"""


class SarpyIOError(SarpyError):
    """A custom exception class for discovered input/output errors."""


############
# module walking to register openers

def check_for_openers(start_package: str, register_method: Callable) -> None:
    """
    Walks the package, and registers the discovered openers. That is, the modules
    with an :meth:`is_a` method.

    Parameters
    ----------
    start_package : str
    register_method : Callable
    """

    module = import_module(start_package)
    for details in pkgutil.walk_packages(module.__path__, start_package+'.'):
        _, module_name, is_pkg = details
        if is_pkg:
            # don't bother checking for packages
            continue
        sub_module = import_module(module_name)
        if hasattr(sub_module, 'is_a'):
            register_method(sub_module.is_a)

    for entry in sarpy._extensions.entry_points(group=start_package):
        sub_module = entry.load()
        if hasattr(sub_module, 'is_a'):
            logger.info(f"Extending {start_package} with {sub_module.__name__}")
            register_method(sub_module.is_a)


#############
# reader implementation for array like data

class BaseReader(object):
    """
    The basic reader definition, using array-like data fetching.

    **Changed in version 1.3.0**
    """

    __slots__ = (
        '_data_segment', '_reader_type', '_closed', '_close_segments',
        '_delete_temp_files')

    def __init__(
            self,
            data_segment: Union[None, DataSegment, Sequence[DataSegment]],
            reader_type: str = 'OTHER',
            close_segments: bool = True,
            delete_files: Optional[Union[str, Sequence[str]]] = None):
        """

        Parameters
        ----------
        data_segment : None|DataSegment|Sequence[DataSegment]
            None is a feasible value for extension tricks, ultimately the data_segments
            must be defined on initialization (by some extension).
        reader_type : str
        close_segments : bool
            Call segment.close() for each data segment on reader.close()?
        delete_files : None|str|Sequence[str]
            Any temp files which should be cleaned up on reader.close()?
            This will occur after closing segments.
        """

        # NB: it's entirely possible under multiple inheritance (class extends
        # two classes each of which extends BaseReader), that this initializer
        # has already been called. Don't override appropriate values, in that case.

        # override regardless here
        reader_type = reader_type.upper()
        if reader_type not in READER_TYPES:
            logger.info(
                'reader_type has value {}, while it is generally expected to be '
                'one of {}'.format(reader_type, READER_TYPES))
        self._reader_type = reader_type

        try:
            _ = self._closed
            # we didn't get an attribute error, so something has already defined
        except AttributeError:
            self._closed = False

        try:
            _ = self._close_segments
        except AttributeError:
            self._close_segments = close_segments

        try:
            _ = self._delete_temp_files
        except AttributeError:
            self._delete_temp_files = []  # type: List[str]

        if delete_files is None:
            pass
        elif isinstance(delete_files, str):
            if delete_files not in self._delete_temp_files:
                self._delete_temp_files.append(delete_files)
        else:
            for entry in delete_files:
                if entry not in self._delete_temp_files:
                    self._delete_temp_files.append(entry)

        try:
            _ = self._data_segment
            # we didn't get an attribute error, so something has already defined it
            self._set_data_segment(data_segment)
            # NB: this will raise a ValueError upon repeated definition attempts.
        except AttributeError:
            self._data_segment = None
            self._set_data_segment(data_segment)

    @property
    def file_name(self) -> Optional[str]:
        """
        None|str: Defined as a convenience property.
        """

        return None

    @property
    def reader_type(self) -> str:
        """
        str: A descriptive string for the type of reader
        """
        return self._reader_type

    @property
    def data_segment(self) -> Union[DataSegment, Tuple[DataSegment, ...]]:
        """
        DataSegment|Tuple[DataSegment, ...]: The data segment collection.
        """

        return self._data_segment

    def _set_data_segment(
            self,
            data_segment: Union[DataSegment, Sequence[DataSegment]]) -> None:
        """
        Sets the data segment collection. This can only be performed once.

        Parameters
        ----------
        data_segment : DataSegment|Sequence[DataSegment]

        Returns
        -------
        None
        """

        if data_segment is None:
            return  # do nothing

        if self._data_segment is not None:
            raise ValueError('data_segment is read only, once set.')

        if isinstance(data_segment, DataSegment):
            data_segment = [data_segment, ]
        if not isinstance(data_segment, Sequence):
            raise TypeError('data_segment must be an instance of DataSegment or a sequence of such instances')

        for entry in data_segment:
            if not isinstance(entry, DataSegment):
                raise TypeError(
                    'Requires all data segment entries to be an instance of DataSegment.\n\t'
                    'Got type {}'.format(type(entry)))
            if not entry.mode == 'r':
                raise ValueError('Each data segment must have mode="r"')

        if len(data_segment) == 1:
            self._data_segment = data_segment[0]
        else:
            self._data_segment = tuple(data_segment)

    @property
    def image_count(self) -> int:
        """
        int: The number of images/data segments from which to read.
        """

        if isinstance(self.data_segment, DataSegment):
            return 1
        else:
            return len(self.data_segment)

    def get_data_segment_as_tuple(self) -> Tuple[DataSegment, ...]:
        """
        Get the data segment collection as a tuple, to avoid the need for redundant
        checking issues.

        Returns
        -------
        Tuple[DataSegment, ...]
        """

        return (self.data_segment, ) if self.image_count == 1 else self._data_segment

    @property
    def data_size(self) -> Union[Tuple[int, ...], Tuple[Tuple[int, ...]]]:
        """
        Tuple[int, ...]|Tuple[Tuple[int, ...], ...]: the output/formatted data
        size(s) of the data segment(s). If there is a single data segment, then
        this will be `Tuple[int, ...]`, otherwise it will be
        `Tuple[Tuple, int, ...], ...]`.
        """

        return self.data_segment.formatted_shape if self.image_count == 1 else \
            tuple(entry.formatted_shape for entry in self.data_segment)

    def get_data_size_as_tuple(self) -> Tuple[Tuple[int, ...], ...]:
        """
        Get the data size collection as a tuple of tuples, to avoid the need
        for redundant checking issues.

        Returns
        -------
        Tuple[Tuple[int, ...], ...]
        """

        return (self.data_size, ) if self.image_count == 1 else self.data_size

    @property
    def raw_data_size(self) -> Union[Tuple[int, ...], Tuple[Tuple[int, ...]]]:
        """
        Tuple[int, ...]|Tuple[Tuple[int, ...], ...]: the raw data size(s) of the
        data segment(s). If there is a single data segment, then this will be
        `Tuple[int, ...]`, otherwise it will be `Tuple[Tuple, int, ...], ...]`.
        """

        return self.data_segment.raw_shape if self.image_count == 1 else \
            tuple(entry.raw_shape for entry in self.data_segment)

    def get_raw_data_size_as_tuple(self) -> Tuple[Tuple[int, ...], ...]:
        """
        Get the raw data size collection as a tuple of tuples, to avoid the need
        for redundant checking issues.

        Returns
        -------
        Tuple[Tuple[int, ...], ...]
        """

        return (self.data_size, ) if self.image_count == 1 else self.data_size

    @property
    def files_to_delete_on_close(self) -> List[str]:
        """
        List[str]: A collection of files to delete on the close operation.
        """

        return self._delete_temp_files

    @property
    def closed(self) -> bool:
        """
        bool: Is the reader closed? Reading will result in a ValueError
        """

        return self._closed

    def _validate_closed(self):
        if not hasattr(self, '_closed') or self._closed:
            raise ValueError('I/O operation of closed writer')

    def read_chip(
            self,
            *ranges: Sequence[Union[None, int, Tuple[int, ...], slice]],
            index: int = 0,
            squeeze: bool = True) -> numpy.ndarray:
        """
        This is identical to :meth:`read`, and presented for backwards compatibility.

        Parameters
        ----------
        ranges : Sequence[Union[None, int, Tuple[int, ...], slice]]
        index : int
        squeeze : bool

        Returns
        -------
        numpy.ndarray

        See Also
        --------
        :meth:`read`.
        """

        return self.__call__(*ranges, index=index, raw=False, squeeze=squeeze)

    def read(
            self,
            *ranges: Union[None, int, Tuple[int, ...], slice],
            index: int = 0,
            squeeze: bool = True) -> numpy.ndarray:
        """
        Read formatted data from the given data segment. Note this is an alias to the
        :meth:`__call__` called as
        :code:`reader(*ranges, index=index, raw=False, squeeze=squeeze)`.

        Parameters
        ----------
        ranges : Sequence[Union[None, int, Tuple[int, ...], slice]]
            The slice definition appropriate for `data_segment[index].read()` usage.
        index : int
            The data_segment index. This is ignored if `image_count== 1`.
        squeeze : bool
            Squeeze length 1 dimensions out of the shape of the return array?

        Returns
        -------
        numpy.ndarray

        See Also
        --------
        See :meth:`sarpy.io.general.data_segment.DataSegment.read`.
        """

        return self.__call__(*ranges, index=index, raw=False, squeeze=squeeze)

    def read_raw(
            self,
            *ranges: Union[None, int, Tuple[int, ...], slice],
            index: int = 0,
            squeeze: bool = True) -> numpy.ndarray:
        """
        Read raw data from the given data segment. Note this is an alias to the
        :meth:`__call__` called as
        :code:`reader(*ranges, index=index, raw=True, squeeze=squeeze)`.

        Parameters
        ----------
        ranges : Sequence[Union[None, int, Tuple[int, ...], slice]]
            The slice definition appropriate for `data_segment[index].read()` usage.
        index : int
            The data_segment index. This is ignored if `image_count== 1`.
        squeeze : bool
            Squeeze length 1 dimensions out of the shape of the return array?

        Returns
        -------
        numpy.ndarray

        See Also
        --------
        See :meth:`sarpy.io.general.data_segment.DataSegment.read_raw`.
        """
        return self.__call__(*ranges, index=index, raw=True, squeeze=squeeze)

    def __call__(
            self,
            *ranges: Union[None, int, Tuple[int, ...], slice],
            index: int = 0,
            raw: bool = False,
            squeeze: bool = True) -> numpy.ndarray:

        self._validate_closed()

        if len(ranges) == 0:
            subscript = None
        else:
            subscript = []
            for rng in ranges:
                if rng is None:
                    subscript.append(slice(None, None, 1))
                elif isinstance(rng, int):
                    subscript.append(slice(rng))
                elif isinstance(rng, tuple):
                    subscript.append(slice(*rng))
                elif isinstance(rng, slice) or rng is Ellipsis:
                    subscript.append(rng)
                else:
                    raise TypeError('Got unexpected type `{}` value for range/slice'.format(type(rng)))
        if isinstance(self._data_segment, tuple):
            ds = self.data_segment[index]
        else:
            ds = self.data_segment

        if raw:
            return ds.read_raw(subscript, squeeze=squeeze)
        else:
            return ds.read(subscript, squeeze=squeeze)

    def __getitem__(self, subscript) -> numpy.ndarray:
        # TODO: document the str usage and index determination

        subscript, string_entries = extract_string_from_subscript(subscript)
        if not isinstance(subscript, (tuple, list)):
            subscript = (subscript, )

        raw = ('raw' in string_entries)
        squeeze = ('nosqueeze' not in string_entries)

        if isinstance(subscript[-1], int):
            the_index = subscript[-1]
            if -self.image_count < the_index < self.image_count:
                return self.__call__(*subscript[:-1], index=subscript[-1], raw=raw, squeeze=squeeze)
        return self.__call__(*subscript, index=0, raw=raw, squeeze=squeeze)

    def close(self) -> None:
        """
        This should perform any necessary clean-up operations, like closing
        open file handles, deleting any temp files, etc.
        """

        if not hasattr(self, '_closed') or self._closed:
            return
        # close all the segments
        if self._close_segments and self._data_segment is not None:
            if isinstance(self._data_segment, DataSegment):
                self._data_segment.close()
            else:
                for entry in self._data_segment:
                    entry.close()
        self._data_segment = None

        # delete temp files
        while len(self._delete_temp_files) > 0:
            filename = self._delete_temp_files.pop()
            try:
                os.remove(filename)
            except FileNotFoundError:
                pass
            except Exception as e:
                logger.error(
                    'Reader attempted to delete temp file {},\n\t'
                    'but got error {}'.format(filename, e))
        self._closed = True

    def __del__(self):
        # NB: this is called when the object is marked for garbage collection
        # (i.e. reference count goes to 0), and the order in which this happens
        # may be unreliable
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

        if exception_type is not None:
            logger.error(
                'The {} file writer generated an exception during processing'.format(
                    self.__class__.__name__))
            # The exception will be reraised.
            # It's unclear how any exception could really be caught.

class FlatReader(BaseReader):
    """
    Class for passing a numpy array straight through as a reader.

    Changed in version 1.3.0
    """

    def __init__(
            self,
            underlying_array: numpy.ndarray,
            reader_type: str = 'OTHER',
            formatted_dtype: Optional[Union[str, numpy.dtype]] = None,
            formatted_shape: Optional[Tuple[int, ...]] = None,
            reverse_axes: Optional[Union[int, Sequence[int]]] = None,
            transpose_axes: Optional[Tuple[int, ...]] = None,
            format_function: Optional[FormatFunction] = None,
            close_segments: bool = True):
        """

        Parameters
        ----------
        underlying_array : numpy.ndarray
        reader_type : str
        formatted_dtype : None|str|numpy.dtype
        formatted_shape : None|Tuple[int, ...]
        reverse_axes : None|Sequence[int]
        transpose_axes : None|Tuple[int, ...]
        format_function : None|FormatFunction
        close_segments : bool
        """

        data_segment = NumpyArraySegment(
            underlying_array, formatted_dtype=formatted_dtype, formatted_shape=formatted_shape,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes,
            format_function=format_function, mode='r')
        BaseReader.__init__(
            self, data_segment, reader_type=reader_type, close_segments=close_segments)


class AggregateReader(BaseReader):
    """
    Aggregate multiple readers into a single reader instance. This default
    aggregate implementation will not preserve any other metadata structures.
    """

    __slots__ = ('_readers', '_index_mapping', '_close_readers')

    def __init__(
            self,
            readers: Sequence[BaseReader],
            reader_type: str = "OTHER",
            close_readers: bool = False):
        """

        Parameters
        ----------
        readers : Sequence[BaseReader]
            The readers.
        reader_type : str
            The reader type string.
        close_readers: bool
            Close all readers on this reader close?
        """

        self._close_readers = close_readers
        self._index_mapping = None
        self._readers = self._validate_readers(readers)
        data_segments = self._define_index_mapping()
        # NB: close_segments is and should be handled by the constituent readers
        BaseReader.__init__(
            self, data_segment=data_segments, reader_type=reader_type, close_segments=False)

    @staticmethod
    def _validate_readers(readers: Sequence[BaseReader]) -> Tuple[BaseReader]:
        """
        Validate the input reader/file collection.

        Parameters
        ----------
        readers : Sequence[BaseReader]

        Returns
        -------
        Tuple[BaseReader]
        """

        if not isinstance(readers, Sequence):
            raise TypeError('input argument must be a sequence of readers. Got type {}'.format(type(readers)))

        # validate each entry
        the_readers = []
        for i, entry in enumerate(readers):
            if not isinstance(entry, BaseReader):
                raise TypeError(
                    'All elements of the input argument must be reader instances. '
                    'Entry {} is of type {}'.format(i, type(entry)))
            the_readers.append(entry)
        return tuple(the_readers)

    def _define_index_mapping(self) -> List[DataSegment]:
        """
        Define the index mapping.

        Returns
        -------
        List[DataSegment]
        """

        # prepare the index mapping workspace
        index_mapping = []

        segments = []
        for i, reader in enumerate(self._readers):
            for j, segment in enumerate(reader.get_data_segment_as_tuple()):
                segments.append(segment)
                index_mapping.append((i, j))
        self._index_mapping = tuple(index_mapping)
        return segments

    @property
    def index_mapping(self) -> Tuple[Tuple[int, int]]:
        """
        Tuple[Tuple[int, int]]: The index mapping of the form (reader index, segment index in reader).
        """

        return self._index_mapping

    def close(self) -> None:
        """
        This should perform any necessary clean-up operations, like closing
        open file handles, deleting any temp files, etc.
        """

        if not hasattr(self, '_closed') or self._closed:
            return

        BaseReader.close(self)
        if self._close_readers and self._readers is not None:
            for entry in self._readers:
                entry.close()
        self._readers = None


#############
# writer implementation for array like data

class BaseWriter(object):
    """
    Writer definition, using array-like data writing.

    Introduced in version 1.3.0
    """

    __slots__ = ('_data_segment', '_closed')

    def __init__(
            self,
            data_segment: Union[DataSegment, Sequence[DataSegment]]):

        self._closed = False
        if isinstance(data_segment, DataSegment):
            data_segment = [data_segment, ]
        if not isinstance(data_segment, Sequence):
            raise TypeError('data_segment must be an instance of DataSegment or a sequence of such instances')

        for entry in data_segment:
            if not isinstance(entry, DataSegment):
                raise TypeError(
                    'Requires all data segment entries to be an instance of DataSegment.\n\t'
                    'Got type {}'.format(type(entry)))
            if not entry.mode == 'w':
                raise ValueError('Each data segment must have mode="w" for writing')

        self._data_segment = tuple(data_segment)

    @property
    def file_name(self) -> Optional[str]:
        """
        None|str: Defined as a convenience property.
        """

        return None

    @property
    def data_segment(self) -> Tuple[DataSegment, ...]:
        """
        Tuple[DataSegment, ...]: The data segment collection.
        """

        return self._data_segment

    @property
    def image_count(self) -> int:
        """
        int: The number of overall images/data segments.
        """

        return len(self.data_segment)

    @property
    def data_size(self) -> Tuple[Tuple[int, ...]]:
        """
        Tuple[Tuple[int, ...], ...]: the formatted data sizes of the data
        segments.
        """

        return tuple(entry.formatted_shape for entry in self.data_segment)

    @property
    def raw_data_size(self) -> Union[Tuple[int, ...], Tuple[Tuple[int, ...]]]:
        """
        Tuple[Tuple[int, ...], ...]: the raw data sizes of the data segments.
        """

        return tuple(entry.raw_shape for entry in self.data_segment)

    @property
    def closed(self) -> bool:
        """
        bool: Is the writer closed? Reading file after writing can
        result in a ValueError if writer was not closed.
        """

        return self._closed

    def _validate_closed(self):
        if not hasattr(self, '_closed') or self._closed:
            raise ValueError('I/O operation of closed writer')

    def write_chip(
            self,
            data: numpy.ndarray,
            start_indices: Optional[Union[int, Tuple[int, ...]]] = None,
            subscript: Optional[Tuple[slice, ...]] = None,
            index: int = 0) -> None:
        """
        This is identical to :meth:`write`, and presented for backwards compatibility.

        Parameters
        ----------
        data : numpy.ndarray
        start_indices : None|int|Tuple[int, ...]
        subscript : None|Tuple[slice, ...]
        index : int

        See Also
        --------
        See :meth:`sarpy.io.general.data_segment.DataSegment.write`.
        """

        self.__call__(data, start_indices=start_indices, subscript=subscript, index=index, raw=False)

    def write(
            self,
            data: numpy.ndarray,
            start_indices: Optional[Union[int, Tuple[int, ...]]] = None,
            subscript: Optional[Tuple[slice, ...]] = None,
            index: int = 0) -> None:
        """
        Write the data to the appropriate data segment. This is an alias to
        :code:`writer(data, start_indices=start_indices, subscript=subscript, index=index, raw=False)`.

        **Only one of `start_indices` and `subscript` should be specified.**

        Parameters
        ----------
        data : numpy.ndarray
            The data to write.
        start_indices : None|int|Tuple[int, ...]
            Assuming a contiguous chunk of data, this provides the starting
            indices of the chunk. Any missing (tail) coordinates will be filled
            in with 0's.
        subscript : None|Tuple[slice, ...]
            In contrast to providing `start_indices`, the slicing definition in
            formatted coordinates pertinent to the specified data segment.
        index : int
            The index of the

        See Also
        --------
        See :meth:`sarpy.io.general.data_segment.DataSegment.write`.
        """

        self.__call__(data, start_indices=start_indices, subscript=subscript, index=index, raw=False)

    def write_raw(
            self,
            data: numpy.ndarray,
            start_indices: Optional[Union[int, Tuple[int, ...]]] = None,
            subscript: Optional[Tuple[slice, ...]] = None,
            index: int = 0) -> None:
        """
        Write the raw data to the file(s). This is an alias to
        :code:`writer(data, start_indices=start_indices, subscript=subscript, index=index, raw=True)`.

        **Only one of `start_indices` and `subscript` should be specified.**

        Parameters
        ----------
        data : numpy.ndarray
            The data to write.
        start_indices : None|int|Tuple[int, ...]
            Assuming a contiguous chunk of data, this provides the starting
            indices of the chunk. Any missing (tail) coordinates will be filled
            in with 0's.
        subscript : None|Tuple[slice, ...]
            In contrast to providing `start_indices`, the slicing definition in
            raw coordinates pertinent to the specified data segment.
        index : int

        See Also
        --------
        See :meth:`sarpy.io.general.data_segment.DataSegment.write_raw`.
        """

        self.__call__(data, start_indices=start_indices, subscript=subscript, index=index, raw=True)

    def __call__(
            self,
            data: numpy.ndarray,
            start_indices: Optional[Union[int, Tuple[int, ...]]] = None,
            subscript: Optional[Tuple[slice, ...]] = None,
            index: int = 0,
            raw: bool = False) -> None:
        """
        Write the data to the given data segment.

        Parameters
        ----------
        data : numpy.ndarray
            The data to write.
        start_indices : None|int|Tuple[int, ...]
            Assuming a contiguous chunk of data, this provides the starting
            indices of the chunk. Any missing (tail) coordinates will be filled
            in with 0's.
        subscript : None|Tuple[slice, ...]
            In contrast to providing `start_indices`, the slicing definition in
            coordinates pertinent to the specified data segment and `raw` value.
        index : int
        raw : bool
        """

        self._validate_closed()
        ds = self.data_segment[index]

        if raw:
            return ds.write_raw(data, start_indices=start_indices, subscript=subscript)
        else:
            if not ds.can_write_regular:
                raise ValueError(
                    'The data segment at index {} can not convert from formatted data to raw data.\n\t'
                    'It is only permitted to use the write_raw() function on this data set,\n\t'
                    'and to write the data in raw (i.e. unformatted) form.')
            return ds.write(data, start_indices=start_indices, subscript=subscript)

    def flush(self, force: bool = False) -> None:
        """
        Try to perform any necessary steps to flush written data to the disk/buffer.

        Parameters
        ----------
        force : bool
            Try force flushing, even for incompletely written data.

        Returns
        -------
        None
        """

        self._validate_closed()
        if self._data_segment is not None:
            for data_segment in self.data_segment:
                data_segment.flush()

    def close(self) -> None:
        """
        This should perform any necessary final steps, like closing
        open file handles, deleting any temp files, etc.
        Trying to read newly created file without closing may raise a ValueError.

        """

        if not hasattr(self, '_closed') or self._closed:
            return

        try:
            # flush the data
            self.flush(force=True)
            # close all the segments
            if self._data_segment is not None:
                for entry in self._data_segment:
                    entry.close()
            self._data_segment = None
            self._closed = True
        except AttributeError:
            self._closed = True
            return

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

        if exception_type is not None:
            logger.error(
                'The {} file writer generated an exception during processing'.format(
                    self.__class__.__name__))
            # The exception will be reraised.
            # It's unclear how any exception could really be caught.



"""
This module provides utilities for attempting to open other image files not
opened by the sicd, sidd, cphd, or crsd reader collections.
"""

import os
from typing import Callable
from sarpy.io.general.base import SarpyIOError, BaseReader, check_for_openers

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


###########
# Module variables
_openers = []
_parsed_openers = False


def register_opener(open_func: Callable) -> None:
    """
    Provide a new opener.

    Parameters
    ----------
    open_func : Callable
        This is required to be a function which takes a single argument (file name).
        This function should return a sarpy.io.general.base.BaseReader instance
        if the referenced file is viable for the underlying type, and None otherwise.

    Returns
    -------
    None
    """

    if not callable(open_func):
        raise TypeError('open_func must be a callable')
    if open_func not in _openers:
        _openers.append(open_func)


def parse_openers() -> None:
    """
    Automatically find the viable openers (i.e. :func:`is_a`) in the various modules.

    Returns
    -------

    """

    global _parsed_openers
    if _parsed_openers:
        return
    _parsed_openers = True

    check_for_openers('sarpy.io.general', register_opener)


def open_general(file_name: str) -> BaseReader:
    """
    Given a file, try to find and return the appropriate reader object.

    Parameters
    ----------
    file_name : str

    Returns
    -------
    BaseReader

    Raises
    ------
    SarpyIOError
    """

    if not os.path.exists(file_name):
        raise SarpyIOError('File {} does not exist.'.format(file_name))
    # parse openers, if not already done
    parse_openers()
    # see if we can find a reader though trial and error
    for opener in _openers:
        reader = opener(file_name)
        if reader is not None:
            return reader

    # If for loop completes, no matching file format was found.
    raise SarpyIOError('Unable to determine image format.')



"""
The reserved extension subheader definitions.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from .base import BaseNITFElement, NITFElement, Unstructured, _IntegerDescriptor,\
    _StringDescriptor, _StringEnumDescriptor, _NITFElementDescriptor, \
    _parse_nitf_element
from .security import NITFSecurityTags, NITFSecurityTags0


class RESUserHeader(Unstructured):
    _size_len = 4


class ReservedExtensionHeader(NITFElement):
    """
    The reserved extension subheader - see standards document Joint BIIF Profile (JBP) for more
    information.
    """

    _ordering = ('RE', 'RESID', 'RESVER', 'Security', 'UserHeader')
    _lengths = {'RE': 2, 'RESID': 25, 'RESVER': 2}
    RE = _StringEnumDescriptor(
        'RE', True, 2, {'RE', }, default_value='RE',
        docstring='File part type.')  # type: str
    RESID = _StringDescriptor(
        'RESID', True, 25, default_value='',
        docstring='Unique RES Type Identifier. This field shall contain a valid alphanumeric '
                  'identifier properly registered with the ISMC.')  # type: str
    RESVER = _IntegerDescriptor(
        'RESVER', True, 2, default_value=1,
        docstring='Version of the Data Definition. This field shall contain the alphanumeric version '
                  'number of the use of the tag. The version number is assigned as part of the '
                  'registration process.')  # type: int
    Security = _NITFElementDescriptor(
        'Security', True, NITFSecurityTags, default_args={},
        docstring='The security tags.')  # type: NITFSecurityTags

    def __init__(self, **kwargs):
        self._RESID = None
        self._UserHeader = None
        super(ReservedExtensionHeader, self).__init__(**kwargs)

    @property
    def UserHeader(self):  # type: () -> RESUserHeader
        """
        RESUserHeader: The RES user header.
        """

        return self._UserHeader

    @UserHeader.setter
    def UserHeader(self, value):
        if not isinstance(value, BaseNITFElement):
            value = _parse_nitf_element(value, RESUserHeader, {}, 'UserHeader', self)
        self._UserHeader = value
        self._load_header_data()

    def _load_header_data(self):
        """
        Load any user defined header specifics.

        Returns
        -------
        None
        """

        pass

    @classmethod
    def _parse_attribute(cls, fields, attribute, value, start):
        if attribute == 'UserHeader':
            val = RESUserHeader.from_bytes(value, start)
            fields['UserHeader'] = val
            return start + val.get_bytes_length()
        else:
            return super(ReservedExtensionHeader, cls)._parse_attribute(fields, attribute, value, start)


class ReservedExtensionHeader0(NITFElement):
    """
    The reserved extension subheader for NITF version 2.0 - see standards
    document MIL-STD-2500A for more information.
    """

    _ordering = ('RE', 'RESID', 'RESVER', 'Security', 'UserHeader')
    _lengths = {'RE': 2, 'RESID': 25, 'RESVER': 2}
    RE = _StringEnumDescriptor(
        'RE', True, 2, {'RE', }, default_value='RE',
        docstring='File part type.')  # type: str
    RESID = _StringDescriptor(
        'RESID', True, 25, default_value='',
        docstring='Unique RES Type Identifier. This field shall contain a valid alphanumeric '
                  'identifier properly registered with the ISMC.')  # type: str
    RESVER = _IntegerDescriptor(
        'RESVER', True, 2, default_value=1,
        docstring='Version of the Data Definition. This field shall contain the alphanumeric version '
                  'number of the use of the tag. The version number is assigned as part of the '
                  'registration process.')  # type: int
    Security = _NITFElementDescriptor(
        'Security', True, NITFSecurityTags0, default_args={},
        docstring='The security tags.')  # type: NITFSecurityTags0

    def __init__(self, **kwargs):
        self._RESID = None
        self._UserHeader = None
        super(ReservedExtensionHeader0, self).__init__(**kwargs)

    @property
    def UserHeader(self):  # type: () -> RESUserHeader
        """
        RESUserHeader: The RES user header.
        """

        return self._UserHeader

    @UserHeader.setter
    def UserHeader(self, value):
        if not isinstance(value, BaseNITFElement):
            value = _parse_nitf_element(value, RESUserHeader, {}, 'UserHeader', self)
        self._UserHeader = value
        self._load_header_data()

    def _load_header_data(self):
        """
        Load any user defined header specifics.

        Returns
        -------
        None
        """

        pass

    @classmethod
    def _parse_attribute(cls, fields, attribute, value, start):
        if attribute == 'UserHeader':
            val = RESUserHeader.from_bytes(value, start)
            fields['UserHeader'] = val
            return start + val.get_bytes_length()
        else:
            return super(ReservedExtensionHeader0, cls)._parse_attribute(fields, attribute, value, start)



"""
Functionality for exporting certain data elements to a kml document
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import zipfile
import logging
import os
import numpy
from typing import Union, List
from uuid import uuid4
from io import BytesIO
from xml.dom import minidom

from sarpy.geometry.geocoords import geodetic_to_ecf, ecf_to_geodetic

try:
    # noinspection PyPackageRequirements
    import PIL
    # noinspection PyPackageRequirements
    import PIL.Image
except ImportError:
    PIL = None

logger = logging.getLogger(__name__)

#################
# default values
_DEFAULT_ICON = 'http://maps.google.com/mapfiles/kml/shapes/shaded_dot.png'


class Document(object):
    """
    The main kml document container, and zip archive if the output file is
    of type kmz. *This is intended to be used as a context manager.*
    """

    __slots__ = ('_doc', '_document', '_archive', '_file', '_closed')

    def __init__(self, file_name=None, **params):
        """

        Parameters
        ----------
        file_name : str|zipfile.ZipFile|file like
            The output location or buffer to which to write the kml/other objects
        params
            The parameters dictionary for file creation.
        """

        self._file = None
        self._archive = None
        self._closed = False
        self._set_file(file_name)

        self._doc = minidom.Document()
        kml = self._doc.createElement('kml')
        self._doc.appendChild(kml)
        kml.setAttribute('xmlns', 'http://www.opengis.net/kml/2.2')
        kml.setAttribute('xmlns:gx', 'http://www.google.com/kml/ext/2.2')
        kml.setAttribute('xmlns:kml', 'http://www.opengis.net/kml/2.2')
        kml.setAttribute('xmlns:atom', 'http://www.w3.org/2005/Atom')
        self._document = self.add_container(kml, 'Document', **params)

    def __str__(self):
        xml = self._doc.toprettyxml(encoding='utf-8')
        if not isinstance(xml, str):
            return xml.decode('utf-8')
        else:
            return xml

    def _set_file(self, file_name):
        if isinstance(file_name, str):
            fext = os.path.splitext(file_name)[1]
            if fext not in ['.kml', '.kmz']:
                logger.warning('file extension should be one of .kml or .kmz, got {}. This will be treated as a kml file.'.format(fext))
            if fext == '.kmz':
                self._archive = zipfile.ZipFile(file_name, 'w', zipfile.ZIP_DEFLATED)
            else:
                self._file = open(file_name, 'w')
        elif isinstance(file_name, zipfile.ZipFile):
            self._archive = file_name
        elif hasattr(file_name, 'write'):
            self._file = file_name
        else:
            raise TypeError('file_name must be a file path, file-like object, or a zipfile.Zipfile instance')

    def close(self):
        if self._closed:
            return

        if self._file is not None:
            self._file.write(str(self))
            self._file.close()
        else:
            self.write_string_to_archive('doc.kml', str(self))
            self._archive.close()
        self._closed = True

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        if exception_type is None:
            self.close()
        else:
            logger.error(
                'The kml/kmz file writer generated an exception during processing.\n\t'
                'Any generated file may be only partially generated and/or corrupt.')
            # The exception will be reraised - it's unclear how any exception could be caught.

    def write_file_to_archive(self, archive_path, file_path):
        """
        Copy the given file into the kmz archive at the given archive location.

        Parameters
        ----------
        archive_path : str
            The location in the archive.
        file_path : str
            The file location on the file system.

        Returns
        -------
        None
        """

        if self._archive is None:
            raise ValueError('No archive defined.')
        self._archive.write(file_path, archive_path, zipfile.ZIP_DEFLATED)

    def write_string_to_archive(self, archive_path, val):
        """
        Write the given string/bytes into the kmz archive at the given location.

        Parameters
        ----------
        archive_path : str
        val : bytes|str

        Returns
        -------
        None
        """

        if self._archive is None:
            raise ValueError('No archive defined.')
        self._archive.writestr(zipfile.ZipInfo(archive_path), val, zipfile.ZIP_DEFLATED)

    def write_image_to_archive(self, archive_path, val, img_format='PNG'):
        """
        Write the given PIL image into the kmz archive at the given location.

        Parameters
        ----------
        archive_path : str
        val : PIL.Image.Image
        img_format : str

        Returns
        -------
        None
        """

        imbuf = BytesIO()
        val.save(imbuf, img_format)
        self.write_string_to_archive(archive_path, imbuf.getvalue())
        imbuf.close()

    # xml node creation elements
    def _create_new_node(self, par, tag):
        # type: (Union[None, minidom.Element], str) -> minidom.Element
        nod = self._doc.createElement(tag)
        if par is None:
            self._document.appendChild(nod)
        else:
            par.appendChild(nod)
        return nod

    def _add_text_node(self, par, tag, value):
        # type: (Union[None, minidom.Element], str, str) -> Union[None, minidom.Element]
        if value is None:
            return None

        nod = self._doc.createElement(tag)
        if isinstance(value, str):
            nod.appendChild(self._doc.createTextNode(value))
        else:
            nod.appendChild(self._doc.createTextNode(str(value)))
        par.appendChild(nod)
        return nod

    def _add_cdata_node(self, par, tag, value):
        # type: (Union[None, minidom.Element], str, str) -> Union[None, minidom.Element]
        if value is None:
            return None

        nod = self._doc.createElement(tag)
        if isinstance(value, str):
            nod.appendChild(self._doc.createCDATASection(value))
        else:
            nod.appendChild(self._doc.createCDATASection(str(value)))
        par.appendChild(nod)
        return nod

    def _add_conditional_text_node(self, par, tag, params, default=None):
        # type: (Union[None, minidom.Element], str, dict, Union[None, str]) -> minidom.Element

        return self._add_text_node(par, tag, params.get(tag, default))

    def _add_conditional_cdata_node(self, par, tag, params, default=None):
        # type: (Union[None, minidom.Element], str, dict, Union[None, str]) -> minidom.Element
        return self._add_cdata_node(par, tag, params.get(tag, default))

    # basic kml container creation
    def add_container(self, par=None, the_type='Placemark', **params):
        """
        For creation of Document, Folder, or Placemark container.

        Parameters
        ----------
        par : None|minidom.Element
        the_type : str
            One of "Placemark", "Folder". The type "Document" can only be used once,
            for the top level construct.
        params
            The dictionary of parameters

        Returns
        -------
        minidom.Element
        """

        if the_type not in ("Placemark", "Folder", "Document"):
            raise ValueError('the_type must be one of ("Placemark", "Folder", "Document")')

        container = self._create_new_node(par, the_type)
        if 'id' in params:
            container.setAttribute('id', params['id'])

        for opt in ['name', 'Snippet', 'styleUrl']:
            self._add_conditional_text_node(container, opt, params)
        self._add_conditional_cdata_node(container, 'description', params)

        # extended data
        if ('schemaUrl' in params) and ('ExtendedData' in params):
            self._add_extended_data(container, **params)
        if ('beginTime' in params) or ('endTime' in params):
            ts = self._create_new_node(container, 'TimeSpan')
            self._add_text_node(ts, 'begin', params.get('beginTime', None))
            self._add_text_node(ts, 'end', params.get('endTime', None))
        elif 'when' in params:
            ts = self._create_new_node(container, 'TimeStamp')
            self._add_text_node(ts, 'when', params['when'])
        elif 'visibility' in params:
            if isinstance(params['visibility'], str):
                visibility = params['visibility']
            else:
                visibility = '1' if params['visibility'] else '0'
            self._add_text_node(container, 'visibility', visibility)
        return container

    # Styles
    def add_style_map(self, style_id, high_id, low_id):
        """
        Creates a styleMap from two given style ids.

        Parameters
        ----------
        style_id : str
        high_id : str
        low_id : str

        Returns
        -------
        None
        """

        sm = self._create_new_node(None, 'StyleMap')
        sm.setAttribute('id', style_id)

        pair1 = self._create_new_node(sm, 'Pair')
        self._add_text_node(pair1, 'key', 'normal')
        self._add_text_node(pair1, 'styleUrl', '#'+low_id)

        pair2 = self._create_new_node(sm, 'Pair')
        self._add_text_node(pair2, 'key', 'highlight')
        self._add_text_node(pair2, 'styleUrl', '#'+high_id)

    def add_style(self, style_id, **params):
        """
        Creates a style for use in the document tree.

        Parameters
        ----------
        style_id : str
            the style id string.
        params
            the dictionary of the parameters

        Returns
        -------
        None
        """

        sty = self._create_new_node(None, 'Style')
        sty.setAttribute('id', style_id)
        if 'line_style' in params:
            self.add_line_style(None, sty, **params['line_style'])
        if 'label_style' in params:
            self.add_label_style(None, sty, **params['label_style'])
        if 'list_style' in params:
            self.add_list_style(None, sty, **params['list_style'])
        if 'icon_style' in params:
            self.add_icon_style(None, sty, **params['icon_style'])
        if 'poly_style' in params:
            self.add_poly_style(None, sty, **params['poly_style'])

    def add_line_style(self, style_id=None, par=None, **params):
        """
        Add line style.

        Parameters
        ----------
        style_id : None|str
            The id, which should not be set if this is a child of a style element.
        par : None|minidom.Element
            The parent node.
        params
            The parameters dictionary.

        Returns
        -------
        None
        """

        sty = self._create_new_node(par, 'LineStyle')
        if style_id is not None:
            sty.setAttribute('id', style_id)
        self._add_conditional_text_node(sty, 'color', params, default='b0ff0000')
        self._add_conditional_text_node(sty, 'width', params, default='1.0')

    def add_list_style(self, style_id=None, par=None, **params):
        """
        Add list style

        Parameters
        ----------
        style_id : None|str
            The id, which should not be set if this is a child of a style element.
        par : None|minidom.Element
            The parent node.
        params
            The parameters dictionary.

        Returns
        -------
        None
        """

        sty = self._create_new_node(par, 'ListStyle')
        if style_id is not None:
            sty.setAttribute('id', style_id)

        item_icon = self._create_new_node(sty, 'ItemIcon')
        self._add_text_node(item_icon, 'href', params.get('icon_ref', _DEFAULT_ICON))

    def add_label_style(self, style_id=None, par=None, **params):
        """
        Add label style

        Parameters
        ----------
        style_id : None|str
            The id, which should not be set if this is a child of a style element.
        par : None|minidom.Element
            The parent node.
        params
            The parameters dictionary.

        Returns
        -------
        None
        """

        sty = self._create_new_node(par, 'LabelStyle')
        if style_id is not None:
            sty.setAttribute('id', style_id)
        self._add_conditional_text_node(sty, 'color', params, default='b0ff0000')
        self._add_conditional_text_node(sty, 'scale', params, default='1.0')

    def add_icon_style(self, style_id=None, par=None, **params):
        """
        Add icon style.

        Parameters
        ----------
        style_id : None|str
            The id, which should not be set if this is a child of a style element.
        par : None|minidom.Element
            The parent node.
        params
            The parameters dictionary.

        Returns
        -------
        None
        """

        sty = self._create_new_node(par, 'IconStyle')
        if style_id is not None:
            sty.setAttribute('id', style_id)
        self._add_conditional_text_node(sty, 'color', params)
        self._add_conditional_text_node(sty, 'scale', params)
        icon = self._create_new_node(sty, 'Icon')
        self._add_text_node(icon, 'href', params.get('icon_ref', _DEFAULT_ICON))

    def add_poly_style(self, style_id=None, par=None, **params):
        """
        Add poly style.

        Parameters
        ----------
        style_id : None|str
            The id, which should not be set if this is a child of a style element.
        par : None|minidom.Element
            The parent node.
        params
            The parameters dictionary.

        Returns
        -------
        None
        """

        sty = self._create_new_node(par, 'PolyStyle')
        if style_id is not None:
            sty.setAttribute('id', style_id)
        self._add_conditional_text_node(sty, 'color', params, default='80ff0000')
        self._add_conditional_text_node(sty, 'fill', params)
        self._add_conditional_text_node(sty, 'outline', params)

    def add_default_style(self):
        """
        Add default style

        The style is created, and appended at root level. The corresponding styleUrl is '#defaultStyle'
        """

        line = {'color': 'ff505050', 'width': '1.0'}
        label = {'color': 'ffc0c0c0', 'scale': '1.0'}
        icon = {'color': 'ffff5050', 'scale': '1.0'}
        poly = {'color': '80ff5050'}
        self.add_style(
            'default_high',
            line_style=line, label_style=label, icon_style=icon, poly_style=poly)

        line['width'] = '0.75'
        label['scale'] = '0.75'
        icon['scale'] = '0.75'
        self.add_style(
            'default_low',
            line_style=line, label_style=label, icon_style=icon, poly_style=poly)
        self.add_style_map('defaultStyle', 'default_high', 'default_low')

    def add_color_ramp(self, colors, high_size=1.0, low_size=0.5, icon_ref=None, name_stem='sty'):
        """
        Adds collection of enumerated styles corresponding to provided array of colors.

        Parameters
        ----------
        colors : numpy.ndarray
            numpy array of shape (N, 4) of 8-bit colors assumed to be RGBA
        high_size : float
            The highlighted size.
        low_size : float
            The regular (low lighted?) size.
        icon_ref : str
            The icon reference.
        name_stem : str
            The string representing the naming convention for the associated styles.

        Returns
        -------
        None
        """

        hline = {'width': 2*high_size}
        hlabel = {'scale': high_size}
        hicon = {'scale': high_size}
        lline = {'width': 2*low_size}
        llabel = {'scale': low_size}
        licon = {'scale': low_size}
        if icon_ref is not None:
            hicon['icon_ref'] = icon_ref
            licon['icon_ref'] = icon_ref
        for i in range(colors.shape[0]):
            col = '{3:02x}{2:02x}{1:02x}{0:02x}'.format(*colors[i, :])
            for di in [hline, hlabel, hicon, lline, llabel, licon]:
                di['color'] = col
            self.add_style(
                '{0!s}{1:d}_high'.format(name_stem, i),
                line_style=hline, label_style=hlabel, icon_style=hicon)
            self.add_style(
                '{0!s}{1:d}_low'.format(name_stem, i),
                line_style=lline, label_style=llabel, icon_style=licon)
            self.add_style_map(
                '{0!s}{1:d}'.format(name_stem, i),
                '{0!s}{1:d}_high'.format(name_stem, i),
                '{0!s}{1:d}_low'.format(name_stem, i))

    # extended data handling
    def add_schema(self, schema_id, field_dict, short_name=None):
        """
        For defining/using the extended data schema capability. **Note that this
        is specifically part of the Google Earth extension of kml, and may not be generally
        supported by anything except Google Earth.**

        Parameters
        ----------
        schema_id : str
            the schema id - must be unique to the id collection document
        field_dict : dict
            dictionary where the key is field name. The corresponding value is a tuple of the form
            `(type, displayName)`, where `displayName` can be `None`. The value of `type` is one of
            the data types permitted:
                * 'string'
                * 'int
                * 'uint'
                * 'short'
                * 'ushort'
                * 'float'
                * 'double'
                * 'bool'
        short_name : None|str
            optional short name for display in the schema

        Returns
        -------
        None
        """

        types = ['string', 'int', 'uint', 'short', 'ushort', 'float', 'double', 'bool']
        sch = self._create_new_node(None, 'Schema')
        sch.setAttribute('id', schema_id)
        if short_name is not None:
            sch.setAttribute('name', short_name)
        for name in field_dict:
            sf = self._doc.createElement('SimpleField')
            typ, dname = field_dict[name]
            sf.setAttribute('name', name)
            if typ in types:
                sf.setAttribute('type', typ)
                sch.appendChild(sf)
            else:
                logger.warning(
                    "Schema '{0!s}' has field '{1!s}' of unrecognized type '{2!s}',\n\t"
                    "which is being excluded.".format(schema_id, name, typ))
            self._add_text_node(sf, 'displayName', dname)

    def _add_extended_data(self, par, **params):
        """
        Adds ExtendedData (schema data) to the parent element. **Note that this
        is specifically part of the Google Earth extension of kml, and may not be generally
        supported by anything except Google Earth.**

        Parameters
        ----------
        par : minidom.Element
        params
            the parameters dictionary

        Returns
        -------
        None
        """

        extended_data = self._create_new_node(par, 'ExtendedData')
        schema_data = self._create_new_node(extended_data, 'SchemaData')
        schema_data.setAttribute('schemaUrl', params['schemaUrl'])
        dat = params['ExtendedData']
        keys = params.get('fieldOrder', sorted(dat.keys()))

        for key in keys:
            # check if data is iterable
            if hasattr(dat[key], '__iter__'):
                array_node = self._create_new_node(schema_data, 'gx:SimpleArrayData')
                array_node.setAttribute('name', key)
                for el in dat[key]:
                    self._add_text_node(array_node, 'gx:value', el)
            else:
                sid = self._add_text_node(schema_data, 'SimpleData', dat[key])
                sid.setAttribute('name', key)

    def add_screen_overlay(self, image_ref, par=None, **params):
        """
        Adds ScreenOverlay object.

        Parameters
        ----------
        image_ref : str
            Reference to appropriate image object, whether in the kmz archive or
            an appropriate url.
        par : None|minidom.Element
            The parent node. Appended at root level if not provided.
        params
            The parameters dictionary.

        Returns
        -------
        minidom.Element
        """

        overlay = self._create_new_node(par, 'ScreenOverlay')
        if 'id' in params:
            overlay.setAttribute('id', params['id'])

        for opt in ['name', 'Snippet', 'styleUrl', 'rotation']:
            self._add_conditional_text_node(overlay, opt, params)
        self._add_conditional_cdata_node(overlay, 'description', params)

        # extended data
        if ('schemaUrl' in params) and ('ExtendedData' in params):
            self._add_extended_data(overlay, **params)

        # overlay parameters
        for opt in ['overlayXY', 'screenXY', 'size', 'rotationXY']:
            olp = self._doc.createElement(opt)
            good = True
            for att in ['x', 'y', 'xunits', 'yunits']:
                key = '{}:{}'.format(opt, att)
                if key in params:
                    olp.setAttribute(att, params[key])
                else:
                    logger.error(
                        'params is missing required key {},\n\t'
                        'so we are aborting screen overlay parameters construction. '
                        'This screen overlay will likely not render correctly.'.format(key))
                    good = False
            if good:
                overlay.appendChild(olp)
        # icon
        ic = self._create_new_node(overlay, 'Icon')
        self._add_text_node(ic, 'href', image_ref)
        return overlay

    # direct kml geometries
    def add_multi_geometry(self, par=None, **params):
        """
        Adds a MultiGeometry object. The MultiGeometry object is really just a container.
        The user must continue adding the primitive Geometry constituents to this container or
        nothing will actually get rendered.

        Parameters
        ----------
        par : None|minidom.Element
            The parent node. If not given, then a Placemark is created.
        params
            The parameters dictionary

        Returns
        -------
        minidom.Element
        """

        if par is None:
            par = self.add_container(**params)
        multigeometry_node = self._create_new_node(par, 'MultiGeometry')
        return multigeometry_node

    def add_polygon(self, outer_coords, inner_coords=None, par=None, condition_coords=True, **params):
        """
        Adds a Polygon element - a polygonal outer region, possibly with polygonal holes removed

        Parameters
        ----------
        outer_coords : str
            comma/space delimited string of coordinates for the outerRing. Format of the string
            :code:`'lon1,lat1,alt1 lon2,lat2,alt2 ...'` with the altitude values optional. If given, the altitude value
            is in meters. The precise interpretation of altitude (relative to the ground, relative to sea level, etc.)
            depends on the value of relevant tags passed down to the LinearRing objects, namely the values for the
            params entries:
                * 'extrude'
                * 'tessellate'
                * 'altitudeMode'
        inner_coords : None|List[str]
            If provided, the coordinates for inner rings.
        par : None|minidom.Element
            The parent node. If not given, then a Placemark is created.
        condition_coords: bool
            If True, the coords are conditioned (longitude-unwrapped and closed) before being added to the document.
            This can help with polygons that span the anti-meridian.
        params
            The parameters dictionary.

        Returns
        -------
        minidom.Element
        """

        def _condition_coords(coord_str):
            coords = numpy.array([[float(v) for v in coord.split(',')] for coord in coord_str.split(' ')])
            if not numpy.array_equal(coords[0], coords[-1]):
                coords = numpy.concatenate((coords, coords[0, numpy.newaxis]))  # first and last coord must match
            # KMZ allows longitude values beyond +/-180 for overlays that overlap the meridian
            coords[:, 0] = numpy.degrees(numpy.unwrap(numpy.radians(coords[:, 0])))
            return " ".join(",".join(str(x) for x in coord) for coord in coords)

        if par is None:
            par = self.add_container(**params)
        polygon_node = self._create_new_node(par, 'Polygon')
        for opt in ['extrude', 'tessellate', 'altitudeMode']:
            self._add_conditional_text_node(polygon_node, opt, params)

        outer_ring_node = self._create_new_node(polygon_node, 'outerBoundaryIs')
        if condition_coords:
            outer_coords = _condition_coords(outer_coords)
        self.add_linear_ring(outer_coords, outer_ring_node)
        if inner_coords is not None:
            for coords in inner_coords:
                inner_ring = self._create_new_node(polygon_node, 'innerBoundaryIs')
                if condition_coords:
                    coords = _condition_coords(coords)
                self.add_linear_ring(coords, inner_ring)

    def add_linear_ring(self, coords, par=None, **params):
        """
        Adds a LinearRing element (closed linear path).

        Parameters
        ----------
        coords : str
            comma/space delimited string of coordinates for the outerRing. Format of the string
            :code:`'lon1,lat1,alt1 lon2,lat2,alt2 ...'` with the altitude values optional. If given, the altitude value
            is in meters. The precise interpretation of altitude (relative to the ground, relative to sea level, etc.)
            depends on the value of relevant tags passed down to the LinearRing objects, namely the values for the
            params entries:
                * 'extrude'
                * 'tessellate'
                * 'altitudeMode'
        par : None|minidom.Element
            The parent node. If not given, then a Placemark is created.
        params
            The parameters dictionary.

        Returns
        -------
        minidom.Element
        """

        if par is None:
            par = self.add_container(**params)
        linear_ring = self._create_new_node(par, 'LinearRing')

        for opt in ['extrude', 'tessellate', 'altitudeMode']:
            self._add_conditional_text_node(linear_ring, opt, params)
        self._add_text_node(linear_ring, 'coordinates', coords)
        return linear_ring

    def add_line_string(self, coords, par=None, **params):
        """
        Adds a LineString element (linear path).

        Parameters
        ----------
        coords : str
            comma/space delimited string of coordinates for the outerRing. Format of the string
            :code:`'lon1,lat1,alt1 lon2,lat2,alt2 ...'` with the altitude values optional. If given, the altitude value
            is in meters. The precise interpretation of altitude (relative to the ground, relative to sea level, etc.)
            depends on the value of relevant tags passed down to the LinearRing objects, namely the values for the
            params entries:
                * 'extrude'
                * 'tessellate'
                * 'altitudeMode'
        par : None|minidom.Element
            The parent node. If not given, then a Placemark is created.
        params
            The parameters dictionary.

        Returns
        -------
        minidom.Element
        """

        if par is None:
            par = self.add_container(**params)
        line_string = self._create_new_node(par, 'LineString')

        for opt in ['extrude', 'tessellate', 'altitudeMode']:
            self._add_conditional_text_node(line_string, opt, params)
        self._add_text_node(line_string, 'coordinates', coords)
        return line_string

    def add_point(self, coords, par=None, **params):
        """
        Adds a Point object.

        Parameters
        ----------
        coords : str
            comma/space delimited string of coordinates for the outerRing. Format of the string
            :code:`'lon1,lat1,alt1 lon2,lat2,alt2 ...'` with the altitude values optional. If given, the altitude value
            is in meters. The precise interpretation of altitude (relative to the ground, relative to sea level, etc.)
            depends on the value of relevant tags passed down to the LinearRing objects, namely the values for the
            params entries:
                * 'extrude'
                * 'tessellate'
                * 'altitudeMode'
        par : None|minidom.Element
            The parent node. If not given, then a Placemark is created.
        params
            The parameters dictionary.

        Returns
        -------
        minidom.Element
        """

        if par is None:
            par = self.add_container(**params)
        point = self._create_new_node(par, 'Point')

        for opt in ['extrude', 'tessellate', 'altitudeMode']:
            self._add_conditional_text_node(point, opt, params)
        self._add_text_node(point, 'coordinates', coords)

    def add_gx_multitrack(self, par=None, **params):
        """
        Adds a MultiTrack from the gx namespace. This is only a container, much like
        a MultiGeometry object, which requires the addition of gx:Track objects. **Note that this
        is specifically part of the Google Earth extension of kml, and may not be generally
        supported by anything except Google Earth.**

        Parameters
        ----------
        par : None|minidom.Element
            The parent node. If not given, then a Placemark is created.
        params
            The parameters dictionary.

        Returns
        -------
            minidom.Element
        """

        if par is None:
            par = self.add_container(**params)
        gx_multitrack = self._create_new_node(par, 'gx:MultiTrack')

        for opt in ['gx:interpolate', 'extrude', 'tessellate', 'altitudeMode']:
            self._add_conditional_text_node(gx_multitrack, opt, params)
        return gx_multitrack

    def add_gx_track(self, coords, whens, angles=None, par=None, **params):
        """
        Adds a Track from the gx namespace. **Note that this
        is specifically part of the Google Earth extension of kml, and may not be generally
        supported by anything except Google Earth.**

        Parameters
        ----------
        coords : List[str]
            list of comma delimited string of coordinates. Format of each string entry: 'lon1,lat1,alt1'
            with altitude values optional. If given, the altitude value is in meters. The precise
            interpretation of altitude (relative to the ground, relative to sea level, etc.) depends on
            the value of relevant tags passed down to the LinearRing objects, namely the values for the
            params entries:
                * 'extrude'
                * 'tessellate'
                * 'altitudeMode'
        whens : List[str]
            list of iso-formatted time strings - entries matching coords
        angles : None|List[str]
            None or list of heading (rotation) angles for the icon. If None, then Google Earth
            infers from the path.
        par : None|minidom.Element
            The parent node. If not given, then a Placemark is created.
        params
            The parameters dictionary.

        Returns
        -------
        minidom.Element
        """

        if par is None:
            par = self.add_container(**params)
        gx_track = self._create_new_node(par, 'gx:Track')
        for opt in ['extrude', 'tessellate', 'altitudeMode']:
            self._add_conditional_text_node(gx_track, opt, params)
        for wh in whens:
            self._add_text_node(gx_track, 'when', wh)
        for coords in coords:
            self._add_text_node(gx_track, 'gx:coord', coords)
        if angles is not None:
            for an in angles:
                self._add_text_node(gx_track, 'gx:angles', '{} 0 0'.format(an))
        if ('ExtendedData' in params) and ('schemaUrl' in params):
            self._add_extended_data(gx_track, **params)
        return gx_track

    def add_ground_overlay(self, image_ref, bounding_box=None, lat_lon_quad=None, par=None, **params):
        """
        Adds GroundOverlay object, defined either from a bounding_box or a lat/lon
        quadrilateral.

        Parameters
        ----------
        image_ref : str
            Reference to appropriate image object, either in the kmz archive or
            an appropriate url.
        bounding_box : None|numpy.ndarray|tuple|list
            list of the form `[latitude max, latitude min, longitude max, longitude min]`
        lat_lon_quad : None|numpy.ndarray|list|tuple
            list of the form [[latitude, longitude]], must have 4 entries. The orientation
            is counter-clockwise from the lower-left image corner.
        par : None|minidom.Element
            The parent node. if not provided, then a Placemark object is created implicitly.
        params
            The parameters dictionary.

        Returns
        -------
        minidom.Element
        """

        if bounding_box is None and lat_lon_quad is None:
            raise ValueError('Either bounding_box or lat_lon_quad must be defined.')
        if bounding_box is not None and lat_lon_quad is not None:
            raise ValueError('Both bounding_box or lat_lon_quad are provided, which is not sensible.')

        if par is None:
            par = self.add_container(**params)
        overlay = self._create_new_node(par, 'GroundOverlay')
        if 'id' in params:
            overlay.setAttribute('id', params['id'])
        for opt in ['name', 'Snippet', 'styleUrl', 'altitude', 'altitudeMode']:
            self._add_conditional_text_node(overlay, opt, params)
        self._add_conditional_cdata_node(overlay, 'description', params)

        # extended data
        if ('schemaUrl' in params) and ('ExtendedData' in params):
            self._add_extended_data(overlay, **params)

        # time parameters
        if ('beginTime' in params) or ('endTime' in params):
            ts = self._create_new_node(overlay, 'TimeSpan')
            self._add_text_node(ts, 'begin', params.get('beginTime', None))
            self._add_text_node(ts, 'end', params.get('endTime', None))
        elif 'when' in params:
            ts = self._create_new_node(overlay, 'TimeStamp')
            self._add_conditional_text_node(ts, 'when', params)

        if bounding_box is not None:
            # latitude/longitude box parameters
            ll = self._create_new_node(overlay, 'LatLonBox')
            for cdir, num in zip(['north', 'south', 'east', 'west'], bounding_box):
                self._add_text_node(ll, cdir, num)
            self._add_conditional_text_node(ll, 'rotation', params)
        elif lat_lon_quad is not None:
            if len(lat_lon_quad) != 4:
                raise ValueError('lat_lon_quad must have length 4.')
            # latitude/longitude quad parameters
            llq = self._create_new_node(overlay, 'gx:LatLonQuad')
            coords = ''
            for entry in lat_lon_quad:
                if isinstance(entry, str):
                    coords += entry.strip() + ' '
                elif len(entry) >= 2:
                    coords += '{0:0.8f},{1:0.8f} '.format(entry[1], entry[0])
                else:
                    raise TypeError('Got unexpected entry type {}'.format(type(entry)))
            self._add_text_node(llq, 'coordinates', coords.strip())

        # icon
        ic = self._create_new_node(overlay, 'Icon')
        self._add_text_node(ic, 'href', image_ref)
        return overlay

    # regionation tools for ground overlays
    def _add_lod(self, par, **params):
        """
        Adds a Level of Detail (LOD) element, which is explicitly a child of Region.

        Parameters
        ----------
        par : minidom.Element
        params

        Returns
        -------
        minidom.Element
        """

        lod = self._create_new_node(par, 'Lod')
        self._add_conditional_text_node(lod, 'minLodPixels', params, '128')
        self._add_conditional_text_node(lod, 'maxLodPixels', params, '-1')
        self._add_conditional_text_node(lod, 'minFadeExtent', params, '0')
        return lod

    def _add_lat_lon_alt_box(self, par, **params):
        """
        Adds LatLonAltBox element, which is explicitly a child of Region.

        Parameters
        ----------
        par : minidom.Element
        params

        Returns
        -------
        minidom.Element
        """

        box = self._create_new_node(par, 'LatLonAltBox')
        for key in ['north', 'south', 'east', 'west', 'minAltitude', 'maxAltitude', 'altitudeMode']:
            self._add_conditional_text_node(box, key, params)
        return box

    def add_region(self, par, **params):
        """
        Adds a region element.

        Parameters
        ----------
        par : None|minidom.Element
        params

        Returns
        -------
        minidom.Element
        """

        reg = self._create_new_node(par, 'Region')
        self._add_lod(reg, **params)
        self._add_lat_lon_alt_box(reg, **params)
        return reg

    def _add_ground_overlay_region_bbox(
            self, image_name, fld, img, image_bounds, bounding_box,
            nominal_image_size, img_format, depth_count=0, **params):
        """
        Helper function for creating a ground overlay region part.

        Parameters
        ----------
        image_name : str
            The image name.
        fld : minidom.Element
        img : PIL.Image.Image
        image_bounds : numpy.ndarray|list|tuple
            Using PIL conventions, of the form `(col min, row min, col max, row max)`.
        bounding_box : numpy.ndarray|tuple|list
            Bounding box of the form `[latitude max, latitude min, longitude max, longitude min]`
        nominal_image_size : int
        img_format : str
        depth_count : int
            What is the depth for our recursion.
        params
            The parameters dictionary.

        Returns
        -------
        None
        """

        col_min, row_min, col_max, row_max = image_bounds

        # determine how to resample this image
        row_length = int(row_max - row_min)
        col_length = int(col_max - col_min)
        cont_recursion = True
        if max(row_length, col_length) < 1.5*nominal_image_size:
            cont_recursion = False
            sample_rows = row_length
            sample_cols = col_length
        elif row_length >= col_length:
            sample_rows = nominal_image_size
            sample_cols = int(col_length*nominal_image_size/float(row_length))
        else:
            sample_cols = nominal_image_size
            sample_rows = int(row_length*nominal_image_size/float(col_length))

        archive_name = 'images/{}.{}'.format(image_name, img_format)
        # resample our image
        pil_box = tuple(int(el) for el in image_bounds)
        this_img = img.crop(pil_box).resize((sample_cols, sample_rows), PIL.Image.LANCZOS)
        self.write_image_to_archive(archive_name, this_img, img_format=img_format)
        # create the ground overlay parameters
        pars = {'name': image_name}
        for key in ['beginTime', 'endTime', 'when']:
            if key in params:
                pars[key] = params[key]
        # create the ground overlay
        gnd_overlay = self.add_ground_overlay(archive_name, bounding_box=bounding_box, par=fld, **pars)
        # add the region
        pars = {}
        if depth_count == 0:
            # root level, no minimum size
            pars['minLodPixels'] = 0
        else:
            pars['minLodPixels'] = 0.3*nominal_image_size
            pars['minFadeExtent'] = 0.3*nominal_image_size
        if cont_recursion:
            pars['maxLodPixels'] = 1.75*nominal_image_size
            pars['maxFadeExtent'] = 0.3*nominal_image_size
        else:
            # leaf, no maximum size
            pars['maxLodPixels'] = -1
        pars['north'] = bounding_box[0]
        pars['south'] = bounding_box[1]
        pars['east'] = bounding_box[2]
        pars['west'] = bounding_box[3]
        self.add_region(gnd_overlay, **pars)

        if cont_recursion:
            # create a list of [(start row, end row)]
            if row_length > 1.5*nominal_image_size:
                split_row = row_min + int(0.5*row_length)
                split_lat = bounding_box[0] + (split_row/float(row_length))*(bounding_box[1] - bounding_box[0])
                row_sizes = [(row_min, split_row), (split_row, row_max)]
                lats = [(bounding_box[0], split_lat), (split_lat, bounding_box[1])]
            else:
                row_sizes = [(row_min, row_max), ]
                lats = [(bounding_box[0], bounding_box[1]), ]

            if col_length > 1.5*nominal_image_size:
                split_col = col_min + int(0.5*col_length)
                split_lon = bounding_box[2] + (split_col/float(row_length))*(bounding_box[3] - bounding_box[2])
                col_sizes = [(col_min, split_col), (split_col, col_max)]
                lons = [(bounding_box[2], split_lon), (split_lon, bounding_box[3])]
            else:
                col_sizes = [(col_min, col_max), ]
                lons = [(bounding_box[2], bounding_box[3]), ]

            count = 0
            for row_bit, lat_bit in zip(row_sizes, lats):
                for col_bit, lon_bit in zip(col_sizes, lons):
                    this_im_name = '{}_{}'.format(image_name, count)
                    this_im_bounds = row_bit + col_bit
                    this_bounding_box = lat_bit + lon_bit
                    self._add_ground_overlay_region_bbox(
                        this_im_name, fld, img, this_im_bounds, this_bounding_box,
                        nominal_image_size, img_format, depth_count=depth_count+1, **params)
                    count += 1

    @staticmethod
    def _split_lat_lon_quad(ll_quad, split_fractions):
        """
        Helper method for recursively splitting the lat/lon quad box.

        Parameters
        ----------
        ll_quad : numpy.ndarray
        split_fractions : list|tuple

        Returns
        -------
        numpy.ndarray
        """

        r1, r2, c1, c2 = split_fractions
        # [0] corresponds to (max_row, 0)
        # [1] corresponds to (max row, max_col)
        # [2] corresponds to (0, max_col)
        # [3] corresponds to (0, 0)

        # do row split
        # [0] = r2*[0] + (1-r2)*[3]
        # [1] = r2*[1] + (1-r2)*[2]
        # [2] = r1*[1] + (1-r1)*[2]
        # [3] = r1*[0] + (1-r1)*[3]
        row_split = numpy.array([
            [r2, 0, 0, 1-r2],
            [0, r2, 1-r2, 0],
            [0, r1, 1-r1, 0],
            [r1, 0, 0, 1-r1],
        ], dtype='float64')

        # do column split
        # [0] = (1-c1)*[0] + c1*[1]
        # [1] = (1-c2)*[0] + c2*[1]
        # [2] = c2*[2] + (1-c2)*[3]
        # [3] = c1*[2] + (1-c1)*[3]
        col_split = numpy.array([
            [1-c1, c1, 0, 0],
            [1-c2, c2, 0, 0],
            [0, 0, c2, 1-c2],
            [0, 0, c1, 1-c1], ], dtype='float64')

        split = col_split.dot(row_split)

        llh_temp = numpy.zeros((4, 3))
        llh_temp[:, :2] = ll_quad
        ecf_coords = geodetic_to_ecf(llh_temp)
        split_ecf = split.dot(ecf_coords)
        return ecf_to_geodetic(split_ecf)[:, :2]

    def _add_ground_overlay_region_quad(
            self, image_name, fld, img, image_bounds, lat_lon_quad,
            nominal_image_size, img_format, depth_count=0, **params):
        """
        Helper function for creating a ground overlay region part.

        Parameters
        ----------
        image_name : str
            The image name.
        fld : minidom.Element
        img : PIL.Image.Image
        image_bounds : numpy.ndarray|list|tuple
            Using PIL conventions, of the form `(col min, row min, col max, row max)`.
        lat_lon_quad : numpy.ndarray
            list of the form [[latitude, longitude]], must have 4 entries.
        nominal_image_size : int
        img_format : str
        depth_count : int
            What is the depth of the recursion?
        params

        Returns
        -------
        None
        """

        bounding_box = [
            float(numpy.max(lat_lon_quad[:, 0])), float(numpy.min(lat_lon_quad[:, 0])),
            float(numpy.max(lat_lon_quad[:, 1])), float(numpy.min(lat_lon_quad[:, 1]))]

        col_min, row_min, col_max, row_max = image_bounds

        # determine how to resample this image
        row_length = int(row_max - row_min)
        col_length = int(col_max - col_min)
        cont_recursion = True
        if max(row_length, col_length) <= 1.5*nominal_image_size:
            cont_recursion = False
            sample_rows = row_length
            sample_cols = col_length
        elif row_length >= col_length:
            sample_rows = nominal_image_size
            sample_cols = int(col_length*nominal_image_size/float(row_length))
        else:
            sample_cols = nominal_image_size
            sample_rows = int(row_length*nominal_image_size/float(col_length))

        logger.info(
            'Processing ({}:{}, {}:{}) into a downsampled image\n\t'
            'of size ({}, {})'.format(
                row_min, row_max, col_min, col_max, sample_rows, sample_cols))

        archive_name = 'images/{}.{}'.format(image_name, img_format)
        pil_box = tuple(int(el) for el in image_bounds)
        # resample our image
        this_img = img.crop(pil_box).resize((sample_cols, sample_rows), PIL.Image.LANCZOS)
        self.write_image_to_archive(archive_name, this_img, img_format=img_format)
        # create the ground overlay parameters
        pars = {'name': image_name}
        for key in ['beginTime', 'endTime', 'when']:
            if key in params:
                pars[key] = params[key]
        # create the ground overlay
        gnd_overlay = self.add_ground_overlay(archive_name, lat_lon_quad=lat_lon_quad, par=fld, **pars)
        # add the region
        pars = {}
        if depth_count == 0:
            # root level, no minimum size
            pars['minLodPixels'] = 0
        else:
            pars['minLodPixels'] = 0.3*nominal_image_size
            pars['minFadeExtent'] = 0.3*nominal_image_size

        if cont_recursion:
            pars['maxLodPixels'] = 1.75*nominal_image_size
            pars['maxFadeExtent'] = 0.3*nominal_image_size
        else:
            # leaf, no maximum size
            pars['maxLodPixels'] = -1

        pars['north'] = bounding_box[0]
        pars['south'] = bounding_box[1]
        pars['east'] = bounding_box[2]
        pars['west'] = bounding_box[3]
        self.add_region(gnd_overlay, **pars)

        if cont_recursion:
            if row_length >= 1.5*nominal_image_size:
                split_row = row_min + int(0.5*row_length)
                row_sizes = [(row_min, split_row), (split_row, row_max)]
            else:
                row_sizes = [(row_min, row_max), ]

            if col_length >= 1.5*nominal_image_size:
                split_col = col_min + int(0.5*col_length)
                col_sizes = [(col_min, split_col), (split_col, col_max)]
            else:
                col_sizes = [(col_min, col_max), ]

            count = 0
            for row_bit in enumerate(row_sizes):
                for col_bit in enumerate(col_sizes):
                    this_im_name = '{}_{}'.format(image_name, count)
                    this_im_bounds = (col_bit[1][0], row_bit[1][0], col_bit[1][1], row_bit[1][1])
                    split_fractions = [
                        (row_bit[1][0] - row_min)/float(row_length),
                        (row_bit[1][1] - row_min)/float(row_length),
                        (col_bit[1][0] - col_min)/float(col_length),
                        (col_bit[1][1] - col_min)/float(col_length)
                    ]

                    this_ll_quad = self._split_lat_lon_quad(lat_lon_quad, split_fractions)

                    self._add_ground_overlay_region_quad(
                        this_im_name, fld, img, this_im_bounds, this_ll_quad,
                        nominal_image_size, img_format, depth_count=depth_count+1,
                        **params)
                    count += 1

    def add_regionated_ground_overlay(
            self, img, par, bounding_box=None, lat_lon_quad=None, img_format='PNG',
            nominal_image_size=1024, **params):
        """
        Adds regionated GroundOverlay objects. This downsamples the image to a pyramid type
        collection of smaller images, and defines the regions. **Requires viable archive.**

        Parameters
        ----------
        img : PIL.Image.Image
            the image instance.
        par : minidom.Element
            the parent node, a folder object will be created and appended to par.
            The overlays will be added below this folder.
        bounding_box : None|numpy.ndarray
            Follows the format for the argument in :func:`add_ground_overlay`.
        lat_lon_quad : None|numpy.ndarray
            Follows the format for the argument in :func:`add_ground_overlay`.
        img_format : str
            string representing a viable Image format. The viable options that will be allowed:
                * 'PNG' - (default) transparency; lossless; good compression
                * 'TIFF' - supports transparency; lossless; poor compression
                * 'JPEG' -  no transparency; lossy; best compression
                * 'GIF' - transparency; lossless; medium compression
            The PIL library actually supports a much larger collection of image formats, but the
            remaining formats are not good candidates for this application.
        nominal_image_size : int
            The nominal image size for splitting. A minimum of 512 will be enforced.
        params
            The parameters dictionary.
        Returns
        -------
        minidom.Element
        """

        nominal_image_size = int(nominal_image_size)
        if nominal_image_size < 512:
            nominal_image_size = 512

        if self._archive is None:
            raise ValueError('We must have a viable archive.')
        if PIL is None:
            raise ImportError(
                'Optional dependency Pillow is required to use this functionality.')
        if not isinstance(img, PIL.Image.Image):
            raise TypeError('We must have that img is a PIL instance, got type {}.'.format(type(img)))

        # validate ground overlay area arguments
        if bounding_box is None and lat_lon_quad is None:
            raise ValueError('Either bounding_box or lat_lon_quad must be defined.')
        if bounding_box is not None and lat_lon_quad is not None:
            raise ValueError('Both bounding_box or lat_lon_quad are provided, which is not sensible.')
        if lat_lon_quad is not None:
            if not isinstance(lat_lon_quad, numpy.ndarray) or lat_lon_quad.ndim != 2 or \
                    lat_lon_quad.shape[0] != 4 or lat_lon_quad.shape[1] != 2:
                raise TypeError('lat_lon_quad, if supplied, must be a numpy array of shape (4, 2).')

        # create our folder object
        fld = self.add_container(par, the_type='Folder', **params)
        # get base name
        base_img_name = '{}-image'.format(uuid4())
        base_img_box = (0, 0, img.size[0], img.size[1])

        if bounding_box is not None:
            self._add_ground_overlay_region_bbox(
                base_img_name, fld, img, base_img_box, bounding_box,
                nominal_image_size, img_format, **params)
        elif lat_lon_quad is not None:
            self._add_ground_overlay_region_quad(
                base_img_name, fld, img, base_img_box, lat_lon_quad,
                nominal_image_size, img_format, **params)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class PTOFF(TREElement):
    def __init__(self, value):
        super(PTOFF, self).__init__()
        self.add_field('IXO', 's', 4, value)
        self.add_field('IYO', 's', 4, value)


class GDPT(TREElement):
    def __init__(self, value, TNUMRD, TNUMCD):
        super(GDPT, self).__init__()
        self.add_field('RCOORD', 's', TNUMRD, value)
        self.add_field('CCOORD', 's', TNUMCD, value)


class GRID(TREElement):
    def __init__(self, value, TNUMRD, TNUMCD):
        super(GRID, self).__init__()
        self.add_field('NXPTS', 'd', 3, value)
        self.add_field('NYPTS', 'd', 3, value)
        if len(value) < self.NXPTS*self.NYPTS*(TNUMRD+TNUMCD):
            raise ValueError(
                'The input string is length {}, but there are supposed '
                'to be {}={}x{} elements each of length {}'.format(
                    len(value), self.NXPTS*self.NYPTS, self.NXPTS, self.NYPTS, TNUMRD+TNUMCD))
        self.add_loop('GDPTs', self.NXPTS*self.NYPTS, GDPT, value, TNUMRD, TNUMCD)


class RSMGGAType(TREElement):
    def __init__(self, value):
        super(RSMGGAType, self).__init__()
        self.add_field('IID', 's', 80, value)
        self.add_field('EDITION', 's', 40, value)
        self.add_field('GGRSN', 'd', 3, value)
        self.add_field('GGCSN', 'd', 3, value)
        self.add_field('GGRFEP', 's', 21, value)
        self.add_field('GGCFEP', 's', 21, value)
        self.add_field('INTORD', 's', 1, value)
        self.add_field('NPLN', 'd', 3, value)
        self.add_field('DELTAZ', 's', 21, value)
        self.add_field('DELTAX', 's', 21, value)
        self.add_field('DELTAY', 's', 21, value)
        self.add_field('ZPLN1', 's', 21, value)
        self.add_field('XIPLN1', 's', 21, value)
        self.add_field('YIPLN1', 's', 21, value)
        self.add_field('REFROW', 's', 9, value)
        self.add_field('REFCOL', 's', 9, value)
        self.add_field('TNUMRD', 'd', 2, value)
        self.add_field('TNUMCD', 'd', 2, value)
        self.add_field('FNUMRD', 'd', 1, value)
        self.add_field('FNUMCD', 'd', 1, value)
        self.add_loop('PTOFFs', self.NPLN-1, PTOFF, value)
        self.add_loop('GRIDs', self.NPLN, GRID, value, self.TNUMRD, self.TNUMCD)


class RSMGGA(TREExtension):
    _tag_value = 'RSMGGA'
    _data_type = RSMGGAType



"""
The reference geometry parameters definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union

import numpy

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import FloatDescriptor, StringEnumDescriptor, \
    SerializableDescriptor
from sarpy.io.complex.sicd_elements.blocks import XYZType

from .base import DEFAULT_STRICT, FLOAT_FORMAT


class SRPType(Serializable):
    """
    The SRP position for the reference vector of the reference channel.
    """

    _fields = ('ECF', 'IAC')
    _required = _fields
    # descriptors
    ECF = SerializableDescriptor(
        'ECF', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='SRP position in ECF coordinates.')  # type: XYZType
    IAC = SerializableDescriptor(
        'IAC', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='SRP position in Image Area Coordinates.')  # type: XYZType

    def __init__(self, ECF=None, IAC=None, **kwargs):
        """

        Parameters
        ----------
        ECF : XYZType|numpy.ndarray|list|tuple
        IAC : XYZType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ECF = ECF
        self.IAC = IAC
        super(SRPType, self).__init__(**kwargs)


class ReferenceGeometryCore(Serializable):
    """
    The base reference geometry implementation.
    """

    _fields = (
        'SideOfTrack', 'SlantRange', 'GroundRange', 'DopplerConeAngle',
        'GrazeAngle', 'IncidenceAngle', 'AzimuthAngle')
    _required = _fields
    _numeric_format = {
        'SlantRange': FLOAT_FORMAT, 'GroundRange': FLOAT_FORMAT, 'DopplerConeAngle': FLOAT_FORMAT,
        'GrazeAngle': FLOAT_FORMAT, 'IncidenceAngle': FLOAT_FORMAT, 'AzimuthAngle': FLOAT_FORMAT}
    # descriptors
    SideOfTrack = StringEnumDescriptor(
        'SideOfTrack', ('L', 'R'), _required, strict=DEFAULT_STRICT,
        docstring='Side of Track parameter for the collection.')  # type: str
    SlantRange = FloatDescriptor(
        'SlantRange', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Slant range from the ARP to the SRP.')  # type: float
    GroundRange = FloatDescriptor(
        'GroundRange', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Ground range from the ARP to the SRP.')  # type: float
    DopplerConeAngle = FloatDescriptor(
        'DopplerConeAngle', _required, strict=DEFAULT_STRICT, bounds=(0, 180),
        docstring='Doppler Cone Angle between ARP velocity and deg SRP Line of '
                  'Sight (LOS).')  # type: float
    GrazeAngle = FloatDescriptor(
        'GrazeAngle', _required, strict=DEFAULT_STRICT, bounds=(0, 90),
        docstring='Grazing angle for the ARP to SRP LOS and the deg Earth Tangent '
                  'Plane (ETP) at the SRP.')  # type: float
    IncidenceAngle = FloatDescriptor(
        'IncidenceAngle', _required, strict=DEFAULT_STRICT, bounds=(0, 90),
        docstring='Incidence angle for the ARP to SRP LOS and the Earth Tangent '
                  'Plane (ETP) at the SRP.')  # type: float
    AzimuthAngle = FloatDescriptor(
        'AzimuthAngle', _required, strict=DEFAULT_STRICT, bounds=(0, 360),
        docstring='Angle from north to the line from the SRP to the ARP ETP '
                  'Nadir (i.e. North to +GPX). Measured clockwise from North '
                  'toward East.')  # type: float

    def __init__(self, SideOfTrack=None, SlantRange=None, GroundRange=None,
                 DopplerConeAngle=None, GrazeAngle=None, IncidenceAngle=None,
                 AzimuthAngle=None, **kwargs):
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.SideOfTrack = SideOfTrack
        self.SlantRange = SlantRange
        self.GroundRange = GroundRange
        self.DopplerConeAngle = DopplerConeAngle
        self.GrazeAngle = GrazeAngle
        self.IncidenceAngle = IncidenceAngle
        self.AzimuthAngle = AzimuthAngle
        super(ReferenceGeometryCore, self).__init__(**kwargs)

    @property
    def look(self):
        """
        int: An integer version of `SideOfTrack`:

            * None if `SideOfTrack` is not defined

            * -1 if SideOfTrack == 'R'

            * 1 if SideOftrack == 'L'
        """

        if self.SideOfTrack is None:
            return None
        return -1 if self.SideOfTrack == 'R' else 1


class MonostaticType(ReferenceGeometryCore):
    """
    Parameters for monostatic collection.
    """

    _fields = (
        'ARPPos', 'ARPVel',
        'SideOfTrack', 'SlantRange', 'GroundRange', 'DopplerConeAngle',
        'GrazeAngle', 'IncidenceAngle', 'AzimuthAngle',
        'TwistAngle', 'SlopeAngle', 'LayoverAngle')
    _required = _fields
    _numeric_format = {
        'SlantRange': FLOAT_FORMAT, 'GroundRange': FLOAT_FORMAT, 'DopplerConeAngle': FLOAT_FORMAT,
        'GrazeAngle': FLOAT_FORMAT, 'IncidenceAngle': FLOAT_FORMAT, 'AzimuthAngle': FLOAT_FORMAT,
        'TwistAngle': FLOAT_FORMAT, 'SlopeAngle': FLOAT_FORMAT, 'LayoverAngle': FLOAT_FORMAT}
    # descriptors
    ARPPos = SerializableDescriptor(
        'ARPPos', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='ARP position in ECF coordinates.')  # type: XYZType
    ARPVel = SerializableDescriptor(
        'ARPVel', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='ARP velocity in ECF coordinates.')  # type: XYZType
    TwistAngle = FloatDescriptor(
        'TwistAngle', _required, strict=DEFAULT_STRICT, bounds=(-90, 90),
        docstring='Twist angle between cross range in the ETP and cross range in '
                  'the slant plane at the SRP.')  # type: float
    SlopeAngle = FloatDescriptor(
        'SlopeAngle', _required, strict=DEFAULT_STRICT, bounds=(0, 90),
        docstring='Angle between the ETP normal (uUP) and the slant plane normal '
                  '(uSPN) at the SRP.')  # type: float
    LayoverAngle = FloatDescriptor(
        'LayoverAngle', _required, strict=DEFAULT_STRICT, bounds=(0, 360),
        docstring='Angle from north to the layover direction in the ETP. Measured '
                  'clockwise from +North toward +East.')  # type: float

    def __init__(self, ARPPos=None, ARPVel=None,
                 SideOfTrack=None, SlantRange=None, GroundRange=None, DopplerConeAngle=None,
                 GrazeAngle=None, IncidenceAngle=None, AzimuthAngle=None,
                 TwistAngle=None, SlopeAngle=None, LayoverAngle=None, **kwargs):
        """

        Parameters
        ----------
        ARPPos : XYZType|numpy.ndarray|list|tuple
        ARPVel : XYZType|numpy.ndarray|list|tuple
        SideOfTrack : float
        SlantRange : float
        GroundRange : float
        DopplerConeAngle : float
        GrazeAngle : float
        IncidenceAngle : float
        AzimuthAngle : float
        TwistAngle : float
        SlopeAngle : float
        LayoverAngle : float
        kwargs
        """
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ARPPos = ARPPos
        self.ARPVel = ARPVel
        self.TwistAngle = TwistAngle
        self.SlopeAngle = SlopeAngle
        self.LayoverAngle = LayoverAngle
        super(MonostaticType, self).__init__(
            SideOfTrack=SideOfTrack, SlantRange=SlantRange, GroundRange=GroundRange,
            DopplerConeAngle=DopplerConeAngle, GrazeAngle=GrazeAngle, IncidenceAngle=IncidenceAngle,
            AzimuthAngle=AzimuthAngle, **kwargs)

    @property
    def MultipathGround(self):
        """
        float: The anticipated angle of multipath features on the ground in degrees.
        """

        if self.TwistAngle is None:
            return None
        else:
            return numpy.rad2deg(
                -numpy.arctan(numpy.tan(numpy.deg2rad(self.TwistAngle)) *
                              numpy.sin(numpy.deg2rad(self.GrazeAngle))))

    @property
    def Multipath(self):
        """
        float: The anticipated angle of multipath features in degrees.
        """
        if self.MultipathGround is None:
            return None
        else:
            return numpy.mod(self.AzimuthAngle - 180 + self.MultipathGround, 360)

    @property
    def Shadow(self):
        """
        float: The anticipated angle of shadow features in degrees.
        """

        return numpy.mod(self.AzimuthAngle - 180, 360)


class BistaticTxRcvType(ReferenceGeometryCore):
    """
    Parameters that describe the Transmit/Receive platforms.
    """

    _fields = (
        'Time', 'Pos', 'Vel',
        'SideOfTrack', 'SlantRange', 'GroundRange', 'DopplerConeAngle',
        'GrazeAngle', 'IncidenceAngle', 'AzimuthAngle')
    _required = _fields
    _numeric_format = {
        'Time': FLOAT_FORMAT, 'SlantRange': FLOAT_FORMAT, 'GroundRange': FLOAT_FORMAT, 'DopplerConeAngle': FLOAT_FORMAT,
        'GrazeAngle': FLOAT_FORMAT, 'IncidenceAngle': FLOAT_FORMAT, 'AzimuthAngle': FLOAT_FORMAT}
    # descriptors
    Time = FloatDescriptor(
        'Time', _required, strict=DEFAULT_STRICT,
        docstring='The transmit or receive time for the vector.')  # type: float
    Pos = SerializableDescriptor(
        'Pos', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='APC position in ECF coordinates.')  # type: XYZType
    Vel = SerializableDescriptor(
        'Vel', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='APC velocity in ECF coordinates.')  # type: XYZType

    def __init__(self, Time=None, Pos=None, Vel=None,
                 SideOfTrack=None, SlantRange=None, GroundRange=None, DopplerConeAngle=None,
                 GrazeAngle=None, IncidenceAngle=None, AzimuthAngle=None, **kwargs):
        """

        Parameters
        ----------
        Time : float
        Pos : XYZType|numpy.ndarray|list|tuple
        Vel : XYZType|numpy.ndarray|list|tuple
        SideOfTrack : float
        SlantRange : float
        GroundRange : float
        DopplerConeAngle : float
        GrazeAngle : float
        IncidenceAngle : float
        AzimuthAngle : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Time = Time
        self.Pos = Pos
        self.Vel = Vel
        super(BistaticTxRcvType, self).__init__(
            SideOfTrack=SideOfTrack, SlantRange=SlantRange, GroundRange=GroundRange,
            DopplerConeAngle=DopplerConeAngle, GrazeAngle=GrazeAngle, IncidenceAngle=IncidenceAngle,
            AzimuthAngle=AzimuthAngle, **kwargs)


class BistaticType(Serializable):
    """

    """

    _fields = (
        'AzimuthAngle', 'AzimuthAngleRate', 'BistaticAngle', 'BistaticAngleRate',
        'GrazeAngle', 'TwistAngle', 'SlopeAngle', 'LayoverAngle',
        'TxPlatform', 'RcvPlatform')
    _required = _fields
    _numeric_format = {
        'AzimuthAngle': FLOAT_FORMAT, 'AzimuthAngleRate': FLOAT_FORMAT, 'BistaticAngle': FLOAT_FORMAT,
        'BistaticAngleRate': FLOAT_FORMAT, 'GrazeAngle': FLOAT_FORMAT, 'TwistAngle': FLOAT_FORMAT,
        'SlopeAngle': FLOAT_FORMAT, 'LayoverAngle': FLOAT_FORMAT}

    # descriptors
    AzimuthAngle = FloatDescriptor(
        'AzimuthAngle', _required, strict=DEFAULT_STRICT, bounds=(0, 360),
        docstring='Angle from north to the projection of the Bistatic pointing vector '
                  '(bP) into the ETP. Measured clockwise from +North toward '
                  '+East.')  # type: float
    AzimuthAngleRate = FloatDescriptor(
        'AzimuthAngleRate', _required, strict=DEFAULT_STRICT,
        docstring='Instantaneous rate of change of the Azimuth Angle '
                  ':math:`d(AZIM)/dt`.')  # type: float
    BistaticAngle = FloatDescriptor(
        'BistaticAngle', _required, strict=DEFAULT_STRICT, bounds=(0, 180),
        docstring='Bistatic angle (Beta) between unit vector from SRP to transmit APC '
                  '(uXmt) and the unit vector from the SRP to the receive '
                  'APC (uRcv).')  # type: float
    BistaticAngleRate = FloatDescriptor(
        'BistaticAngleRate', _required, strict=DEFAULT_STRICT,
        docstring='Instantaneous rate of change of the bistatic angle '
                  ':math:`d(Beta)/dt)`.')  # type: float
    GrazeAngle = FloatDescriptor(
        'GrazeAngle', _required, strict=DEFAULT_STRICT, bounds=(0, 90),
        docstring='Angle between the bistatic pointing vector and the ETP at the '
                  'SRP.')  # type: float
    TwistAngle = FloatDescriptor(
        'TwistAngle', _required, strict=DEFAULT_STRICT, bounds=(-90, 90),
        docstring='Angle between cross range in the ETP at the SRP and cross range '
                  'in the instantaneous plane of maximum bistatic resolution. '
                  'Note - For monostatic imaging, the plane of maximum resolution is '
                  'the instantaneous slant plane.')  # type: float
    SlopeAngle = FloatDescriptor(
        'SlopeAngle', _required, strict=DEFAULT_STRICT, bounds=(0, 90),
        docstring='Angle between the ETP normal and the normal to the instantaneous '
                  'plane of maximum bistatic resolution.')  # type: float
    LayoverAngle = FloatDescriptor(
        'LayoverAngle', _required, strict=DEFAULT_STRICT, bounds=(0, 360),
        docstring='Angle from north to the bistatic layover direction in the ETP. '
                  'Measured clockwise from +North toward +East.')  # type: float
    TxPlatform = SerializableDescriptor(
        'TxPlatform', BistaticTxRcvType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the Transmit platform.')  # type: BistaticTxRcvType
    RcvPlatform = SerializableDescriptor(
        'RcvPlatform', BistaticTxRcvType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the Receive platform.')  # type: BistaticTxRcvType

    def __init__(self, AzimuthAngle=None, AzimuthAngleRate=None, BistaticAngle=None,
                 BistaticAngleRate=None, GrazeAngle=None, TwistAngle=None,
                 SlopeAngle=None, LayoverAngle=None, TxPlatform=None,
                 RcvPlatform=None, **kwargs):
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.AzimuthAngle = AzimuthAngle
        self.AzimuthAngleRate = AzimuthAngleRate
        self.BistaticAngle = BistaticAngle
        self.BistaticAngleRate = BistaticAngleRate
        self.GrazeAngle = GrazeAngle
        self.TwistAngle = TwistAngle
        self.SlopeAngle = SlopeAngle
        self.LayoverAngle = LayoverAngle
        self.TxPlatform = TxPlatform
        self.RcvPlatform = RcvPlatform
        super(BistaticType, self).__init__(**kwargs)


class ReferenceGeometryType(Serializable):
    """
    Parameters that describe the collection geometry for the reference vector
    of the reference channel.
    """

    _fields = ('SRP', 'ReferenceTime', 'SRPCODTime', 'SRPDwellTime', 'Monostatic', 'Bistatic')
    _required = ('SRP', 'ReferenceTime', 'SRPCODTime', 'SRPDwellTime')
    _choice = ({'required': True, 'collection': ('Monostatic', 'Bistatic')}, )
    _numeric_format = {'ReferenceTime': FLOAT_FORMAT, 'SRPCODTime': FLOAT_FORMAT, 'SRPDwellTime': FLOAT_FORMAT}
    # descriptors
    SRP = SerializableDescriptor(
        'SRP', SRPType, _required, strict=DEFAULT_STRICT,
        docstring='The SRP position for the reference vector of the reference '
                  'channel.')  # type: SRPType
    ReferenceTime = FloatDescriptor(
        'ReferenceTime', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Reference time for the selected reference vector, in '
                  'seconds.')  # type: float
    SRPCODTime = FloatDescriptor(
        'SRPCODTime', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='The COD Time for point on the reference surface, in '
                  'seconds.')  # type: float
    SRPDwellTime = FloatDescriptor(
        'SRPDwellTime', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='')  # type: float
    Monostatic = SerializableDescriptor(
        'Monostatic', MonostaticType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters for monstatic collection.')  # type: Union[None, MonostaticType]
    Bistatic = SerializableDescriptor(
        'Bistatic', BistaticType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters for bistatic collection.')  # type: Union[None, BistaticType]

    def __init__(self, SRP=None, ReferenceTime=None, SRPCODTime=None, SRPDwellTime=None,
                 Monostatic=None, Bistatic=None, **kwargs):
        """

        Parameters
        ----------
        SRP : SRPType
        ReferenceTime : float
        SRPCODTime : float
        SRPDwellTime : float
        Monostatic : None|MonostaticType
        Bistatic : None|BistaticType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.SRP = SRP
        self.ReferenceTime = ReferenceTime
        self.SRPCODTime = SRPCODTime
        self.SRPDwellTime = SRPDwellTime
        self.Monostatic = Monostatic
        self.Bistatic = Bistatic
        super(ReferenceGeometryType, self).__init__(**kwargs)



"""
Basic image registration tools
"""

__classification__ = "UNCLASSIFIED"
__author__ = 'Thomas McCullough'

import logging
from typing import Sequence, Union, Tuple, Any

import numpy
from scipy.optimize import minimize

from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.product.sidd1_elements.SIDD import SIDDType as SIDDType1
from sarpy.io.product.sidd2_elements.SIDD import SIDDType as SIDDType2
from sarpy.geometry.geocoords import ecf_to_geodetic
from sarpy.geometry.point_projection import ground_to_image


logger = logging.getLogger(__name__)


def best_physical_location_fit(
        structs: Sequence[Union[SICDType, SIDDType1, SIDDType2]],
        locs: Union[numpy.ndarray, list, tuple],
        **minimization_args) -> Tuple[numpy.ndarray, float, Any]:
    """
    Given a collection of SICD and/or SIDDs and a collection of image coordinates, 
    each of which identifies the pixel location of the same feature in the 
    respective image, determine the (best fit) geophysical location of this feature. 

    This assumes that any adjustable parameters used for the SICD/SIDD projection 
    model have already been applied (via :func:`define_coa_projection`).

    Parameters
    ----------
    structs : Sequence[SICDType|SIDDType1|SIDDType2]
        The collection of sicds/sidds, of length `N`
    locs : numpy.ndarray|list|tuple
        The image coordinate collection, of shape `(N, 2)`
    minimization_args
        The keyword arguments (after `args` argument) passed through to
        :func:`scipy.optimize.minimize`. This will default to `'Powell'` 
        optimization, which seems generally much more reliable for this 
        problem than the steepest descent based approaches.

    Returns
    -------
    ecf_location : numpy.ndarray
        The location in question, in ECF coordinates
    residue : float
        The mean square residue of the physical distance between the given
        location and the image locations projected into the surface of
        given HAE value
    result
        The minimization result object
    """

    def get_mean_location(hae_value, log_residue=False):
        ecf_locs = numpy.zeros((points, 3), dtype='float64')
        for i, (loc, struct) in enumerate(zip(locs, structs)):
            ecf_locs[i, :] = struct.project_image_to_ground(loc, projection_type='HAE', hae0=hae_value)
        ecf_mean = numpy.mean(ecf_locs, axis=0)
        diff = ecf_locs - ecf_mean
        residue = numpy.sum(diff*diff, axis=1)
        if log_residue:
            logger.info(
                'best physical location residues [m^2]\n{}'.format(residue))

        avg_residue = numpy.mean(residue)
        return ecf_mean, avg_residue

    def average_residue(hae_value):
        return get_mean_location(hae_value)[1]

    points = len(structs)
    if points < 2:
        raise ValueError(
            'At least 2 structs must be present to determine the best location')
    if points != locs.shape[0]:
        raise ValueError(
            'The number of structs must match the number of locations')

    struct0 = structs[0]
    if isinstance(struct0, SICDType):
        h0 = struct0.GeoData.SCP.LLH.HAE
    elif isinstance(struct0, (SIDDType1, SIDDType2)):
        ref_ecf = struct0.Measurement.ReferencePoint.ECEF.get_array()
        h0 = ecf_to_geodetic(ref_ecf)[2]
    else:
        raise TypeError('Got unexpected structure type {}'.format(type(struct0)))

    if 'method' not in minimization_args:
        minimization_args['method'] = 'Powell'

    result = minimize(average_residue, h0, **minimization_args)
    if not result.success:
        raise ValueError('Optimization failed {}'.format(result))

    values = get_mean_location(result.x, log_residue=True)
    return values[0], values[1], result


def _find_best_adjustable_parameters_sicd(
        sicd: SICDType,
        ecf_coords: numpy.ndarray,
        img_coords: numpy.ndarray,
        **minimization_args) -> Tuple[numpy.ndarray, numpy.ndarray, float, float, Any]:
    """
    Find the best projection model adjustable parameters (in `'ECF'` coordinate frame)
    to fit the geophysical coordinate locations to the image coordinate locations.

    Parameters
    ----------
    sicd : SICDType
    ecf_coords : numpy.ndarray
        The geophysical coordinates of shape `(N, 3)` for the identified features
    img_coords : numpy.ndarray
        The image coordinates of shape `(N, 2)` for the identified features
    minimization_args
        The keyword arguments (after `args` argument) passed through to
        :func:`scipy.optimize.minimize`. This will default to `'Powell'`
        optimization, which seems generally much more reliable for this
        problem than the steepest descent based approaches.

    Returns
    -------
    delta_arp : numpy.ndarray
    delta_varp : numpy.ndarray
    delta_range : float
    residue : float
        Average pixel coordinate distance across the features between projected
        and observed pixel locations
    result
        The minimization result
    """

    row = sicd.Grid.Row.UVectECF.get_array()
    col = sicd.Grid.Col.UVectECF.get_array()

    def get_params(perturb):
        da = perturb[0]*row + perturb[1]*col
        dv = perturb[2:5]
        dr = perturb[5]
        return da, dv, dr

    def get_sq_residue(perturb):
        da, dv, dr = get_params(perturb)
        img_proj, _, _ = ground_to_image(
            ecf_coords, sicd, max_iterations=100, use_structure_coa=False,
            delta_arp=da, delta_varp=dv, range_bias=dr,
            adj_params_frame='ECF')
        diff = (img_proj - img_coords)
        return numpy.sum(diff*diff, axis=1)

    def average_residue(perturb):
        res = get_sq_residue(perturb)
        return numpy.mean(res)

    if 'method' not in minimization_args:
        minimization_args['method'] = 'Powell'

    p0 = numpy.zeros((6, ), dtype='float64')
    result = minimize(average_residue, p0, **minimization_args)
    if not result.success:
        raise ValueError('Optimization failed {}'.format(result))

    logger.info(
        'best adjustable parameters residues [pix^2]\n{}'.format(
            get_sq_residue(result.x)))

    delta_arp, delta_varp, delta_range = get_params(result.x)
    return delta_arp, delta_varp, delta_range, result.fun, result


def _find_best_adjustable_parameters(
        struct: Union[SICDType, SIDDType1, SIDDType2],
        ecf_coords: numpy.ndarray,
        img_coords: numpy.ndarray,
        **minimization_args) -> Tuple[numpy.ndarray, numpy.ndarray, float, float, Any]:
    """
    Find the best projection model adjustable parameters (in `'ECF'` coordinate frame)
    to fit the geophysical coordinate locations to the image coordinate locations.

    Parameters
    ----------
    struct : SICDType|SIDDType1|SIDDType2
    ecf_coords : numpy.ndarray
        The geophysical coordinates of shape `(N, 3)` for the identified features
    img_coords : numpy.ndarray
        The image coordinates of shape `(N, 2)` for the identified features
    minimization_args
        The keyword arguments (after `args` argument) passed through to
        :func:`scipy.optimize.minimize`. This will default to `'Powell'`
        optimization, which seems generally much more reliable for this
        problem than the steepest descent based approaches.

    Returns
    -------
    delta_arp : numpy.ndarray
    delta_varp : numpy.ndarray
    delta_range : float
    residue : float
        Average pixel coordinate distance across the features between projected
        and observed pixel locations
    result
        The minimization result
    """

    def get_params(perturb):
        da = perturb[0:3]
        dv = perturb[3:6]
        dr = perturb[6]
        return da, dv, dr

    def get_sq_residue(perturb):
        da, dv, dr = get_params(perturb)
        img_proj, _, _ = ground_to_image(
            ecf_coords, struct, max_iterations=100, use_structure_coa=False,
            delta_arp=da, delta_varp=dv, range_bias=dr,
            adj_params_frame='ECF')
        diff = (img_proj - img_coords)
        return numpy.sum(diff*diff, axis=1)

    def average_residue(perturb):
        return numpy.mean(get_sq_residue(perturb))

    if 'method' not in minimization_args:
        minimization_args['method'] = 'Powell'

    p0 = numpy.zeros((7,), dtype='float64')
    result = minimize(average_residue, p0, **minimization_args)
    if not result.success:
        raise ValueError('Optimization failed {}'.format(result))

    logger.info(
        'best adjustable parameters residues [pix^2]\n{}'.format(
            get_sq_residue(result.x)))

    delta_arp, delta_varp, delta_range = get_params(result.x)
    return delta_arp, delta_varp, delta_range, result.fun, result


def find_best_adjustable_parameters(
        struct: Union[SICDType, SIDDType1, SIDDType2],
        ecf_coords: numpy.ndarray,
        img_coords: numpy.ndarray,
        **minimization_args) -> Tuple[numpy.ndarray, numpy.ndarray, float, float, Any]:
    """
    Find the best projection model adjustable parameters (in `'ECF'` coordinate frame)
    to fit the geophysical coordinate locations to the image coordinate locations.

    Parameters
    ----------
    struct : SICDType|SIDDType1|SIDDType2
    ecf_coords : numpy.ndarray
        The geophysical coordinates of shape `(N, 3)` for the identified features
    img_coords : numpy.ndarray
        The image coordinates of shape `(N, 2)` for the identified features
    minimization_args
        The keyword arguments (after `args` argument) passed through to
        :func:`scipy.optimize.minimize`. This will default to `'Powell'`
        optimization, which seems generally much more reliable for this
        problem than the steepest descent based approaches.

    Returns
    -------
    delta_arp : numpy.ndarray
    delta_varp : numpy.ndarray
    delta_range : float
    residue : float
        Average pixel coordinate distance across the features between projected
        and observed pixel locations
    result
        The minimization result
    """

    return _find_best_adjustable_parameters(struct, ecf_coords, img_coords, **minimization_args)
